"""Providing a multi-objective process without logging with a single space."""

from math import isfinite
from typing import Any, Callable, Final, cast

import numpy as np
from numpy import copyto
from pycommons.io.path import Path
from pycommons.types import check_int_range, type_error

from moptipy.api._process_base import _TIME_IN_NS, _ns_to_ms, _ProcessBase
from moptipy.api.algorithm import Algorithm
from moptipy.api.logging import (
    KEY_ARCHIVE_F,
    KEY_ARCHIVE_MAX_SIZE,
    KEY_ARCHIVE_PRUNE_LIMIT,
    KEY_ARCHIVE_SIZE,
    KEY_BEST_FS,
    PREFIX_SECTION_ARCHIVE,
    PROGRESS_CURRENT_F,
    PROGRESS_FES,
    PROGRESS_TIME_MILLIS,
    SCOPE_PRUNER,
    SECTION_ARCHIVE_QUALITY,
    SECTION_PROGRESS,
    SUFFIX_SECTION_ARCHIVE_Y,
)
from moptipy.api.mo_archive import (
    MOArchivePruner,
    MORecord,
    check_mo_archive_pruner,
)
from moptipy.api.mo_problem import MOProblem
from moptipy.api.mo_process import MOProcess
from moptipy.api.space import Space
from moptipy.utils.logger import KeyValueLogSection, Logger
from moptipy.utils.nputils import array_to_str, np_to_py_number


class _MOProcessNoSS(MOProcess, _ProcessBase):
    """
    An internal class process implementation.

    This class implements a stand-alone process without explicit logging where
    the search and solution space are the same.
    """

    def __init__(self,
                 solution_space: Space,
                 objective: MOProblem,
                 algorithm: Algorithm,
                 pruner: MOArchivePruner,
                 archive_max_size: int,
                 archive_prune_limit: int,
                 log_file: Path | None = None,
                 rand_seed: int | None = None,
                 max_fes: int | None = None,
                 max_time_millis: int | None = None,
                 goal_f: int | float | None = None) -> None:
        """
        Perform the internal initialization. Do not call directly.

        :param solution_space: the search- and solution space.
        :param objective: the objective function
        :param algorithm: the optimization algorithm
        :param pruner: the archive pruner
        :param archive_max_size: the maximum archive size after pruning
        :param archive_prune_limit: the archive size above which pruning will
            be performed
        :param log_file: the optional log file
        :param rand_seed: the optional random seed
        :param max_fes: the maximum permitted function evaluations
        :param max_time_millis: the maximum runtime in milliseconds
        :param goal_f: the goal objective value. if it is reached, the process
            is terminated
        """
        _ProcessBase.__init__(
            self, solution_space, objective, algorithm, log_file, rand_seed,
            max_fes, max_time_millis, goal_f)
        self.f_dimension = objective.f_dimension  # type: ignore
        self.f_create = objective.f_create  # type: ignore
        self.f_validate = objective.f_validate  # type: ignore
        self.f_dtype = objective.f_dtype  # type: ignore
        self.f_dominates = objective.f_dominates  # type: ignore

        #: the internal evaluation function
        self._f_evaluate: Final[Callable[
            [Any, np.ndarray], int | float]] = objective.f_evaluate

        #: the temporary variable for objective function evaluations
        self._fs_temp: Final[np.ndarray] = self.f_create()
        #: the holder for the objective vector of the current best solution
        self._current_best_fs: Final[np.ndarray] = self.f_create()
        #: the internal archive pruner
        self._pruner: Final[MOArchivePruner] = check_mo_archive_pruner(pruner)
        #: the fast call to the pruning routine
        self._prune: Final[Callable[[list[MORecord], int, int], None]] \
            = pruner.prune
        #: the archive prune limit
        self._archive_prune_limit: Final[int] = check_int_range(
            archive_prune_limit, "archive_prune_limit")
        #: the maximum archive size
        self._archive_max_size: Final[int] = check_int_range(
            archive_max_size, "archive_max_size", 1, archive_prune_limit)
        #: the current archive size
        self._archive_size: int = 0
        #: the internal archive (pre-allocated to the prune limit)
        self._archive: Final[list[MORecord]] = []

    def _after_init(self) -> None:
        self._archive.extend(
            MORecord(self.create(), self.f_create())
            for _ in range(self._archive_prune_limit))
        super()._after_init()

    def check_in(self, x: Any, fs: np.ndarray,
                 prune_if_necessary: bool = False) -> bool:
        """
        Check a solution into the archive.

        :param x: the point in the search space
        :param fs: the vector of objective values
        :param prune_if_necessary: should we prune the archive if it becomes
            too large? `False` means that the archive may grow unbounded
        :returns: `True` if the solution was non-dominated, `False` if it was
            dominated by at least one solution in the archive
        """
        archive: Final[list[MORecord]] = self._archive
        added_to_archive: bool = False
        archive_size: int = self._archive_size
        # we update the archive
        domination: Final[Callable[[np.ndarray, np.ndarray], int]] \
            = self.f_dominates
        for i in range(archive_size - 1, -1, -1):
            ae: MORecord = archive[i]
            d: int = domination(fs, ae.fs)
            if d < 0:  # the new solution dominates an archived one
                if added_to_archive:  # if already added, shrink archive
                    archive_size = archive_size - 1
                    archive[archive_size], archive[i] = \
                        ae, archive[archive_size]
                else:  # if not added, overwrite dominated solution
                    self.copy(ae.x, x)
                    copyto(ae.fs, fs)
                    added_to_archive = True
            elif d > 0:
                return False

        if added_to_archive:  # already added, can quit
            self._archive_size = archive_size
        else:  # still need to add
            if archive_size >= len(archive):
                ae = MORecord(self.create(), self.f_create())
                archive.append(ae)
            else:
                ae = archive[archive_size]
            self.copy(ae.x, x)
            copyto(ae.fs, fs)
            archive_size += 1
            if prune_if_necessary \
                    and (archive_size > self._archive_prune_limit):
                self._prune(archive, self._archive_max_size, archive_size)
                self._archive_size = self._archive_max_size
            else:
                self._archive_size = archive_size
        return True

    def f_evaluate(self, x, fs: np.ndarray) -> float | int:
        if self._terminated:
            if self._knows_that_terminated:
                raise ValueError("The process has been terminated and "
                                 "the algorithm knows it.")
            return self._current_best_f

        result: Final[int | float] = self._f_evaluate(x, fs)
        self._current_fes = current_fes = self._current_fes + 1
        do_term: bool = current_fes >= self._end_fes

        improved: bool = False
        if result < self._current_best_f:
            improved = True
            self._current_best_f = result
            copyto(self._current_best_fs, fs)
            self._copy_y(self._current_best_y, x)
            do_term = do_term or (result <= self._end_f)

        if self.check_in(x, fs, True) or improved:
            self._last_improvement_fe = current_fes
            self._current_time_nanos = ctn = _TIME_IN_NS()
            self._last_improvement_time_nanos = ctn

        if do_term:
            self.terminate()

        return result

    def evaluate(self, x) -> float | int:
        return self.f_evaluate(x, self._fs_temp)

    def register(self, x, f: int | float) -> None:
        raise ValueError(
            "register is not supported in multi-objective optimization")

    def get_archive(self) -> list[MORecord]:
        return self._archive[0:self._archive_size]

    def get_copy_of_best_fs(self, fs: np.ndarray) -> None:
        if self._current_fes > 0:
            return copyto(fs, self._current_best_fs)
        raise ValueError("No current best available.")

    def _log_own_parameters(self, logger: KeyValueLogSection) -> None:
        super()._log_own_parameters(logger)
        logger.key_value(KEY_ARCHIVE_MAX_SIZE, self._archive_max_size)
        logger.key_value(KEY_ARCHIVE_PRUNE_LIMIT, self._archive_prune_limit)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        _ProcessBase.log_parameters_to(self, logger)
        with logger.scope(SCOPE_PRUNER) as sc:
            self._pruner.log_parameters_to(sc)

    def _log_best(self, kv: KeyValueLogSection) -> None:
        super()._log_best(kv)
        kv.key_value(KEY_BEST_FS, array_to_str(self._current_best_fs))
        kv.key_value(KEY_ARCHIVE_SIZE, self._archive_size)

    def _log_and_check_archive_entry(self, index: int, rec: MORecord,
                                     logger: Logger) -> int | float:
        """
        Write an archive entry.

        :param index: the index of the entry
        :param rec: the record to verify
        :param logger: the logger
        :returns: the objective value
        """
        self.f_validate(rec.fs)
        self.validate(rec.x)
        tfs: Final[np.ndarray] = self._fs_temp
        f: Final[int | float] = self._f_evaluate(rec.x, tfs)
        if not np.array_equal(tfs, rec.fs):
            raise ValueError(
                f"expected {rec.fs} but got {tfs} when re-evaluating {rec}")
        if not isinstance(f, int | float):
            raise type_error(f, "scalarized objective value", (int, float))
        if not isfinite(f):
            raise ValueError(f"scalarized objective value {f} is not finite")

        with logger.text(f"{PREFIX_SECTION_ARCHIVE}{index}"
                         f"{SUFFIX_SECTION_ARCHIVE_Y}") as lg:
            lg.write(self.to_str(rec.x))
        return f

    def _write_log(self, logger: Logger) -> None:
        super()._write_log(logger)

        if self._archive_size > 0:
            # write and verify the archive
            archive: Final[list[MORecord]] = \
                self._archive[0:self._archive_size]
            archive.sort()
            qualities: Final[list[list[int | float]]] = []
            for i, rec in enumerate(archive):
                q: list[int | float] = [
                    np_to_py_number(n) for n in rec.fs]
                q.insert(0, self._log_and_check_archive_entry(i, rec, logger))
                qualities.append(q)

            # now write the qualities
            headline: list[str] = [
                f"{KEY_ARCHIVE_F}{i}" for i in range(self.f_dimension())]
            headline.insert(0, KEY_ARCHIVE_F)
            with logger.csv(SECTION_ARCHIVE_QUALITY, headline) as csv:
                for qq in qualities:
                    csv.row(qq)

    def _write_mo_log(self,
                      log: list[list[int | float | np.ndarray]],
                      start_time: int,
                      keep_all: bool,
                      logger: Logger) -> None:
        """
        Write the multi-objective log to the logger.

        :param log: the log
        :param start_time: the start time
        :param keep_all: do we need to keep all entries?
        :param logger: the destination logger
        """
        loglen = len(log)
        if loglen <= 0:
            return

        domination: Final[Callable[[np.ndarray, np.ndarray], int]] \
            = self.f_dominates

        if not keep_all:
            # first we clean the log from potentially dominated entries
            for i in range(loglen - 1, 0, -1):
                reci = log[i]
                fi = cast(int | float, reci[2])
                fsi = cast(np.ndarray, reci[3])
                for j in range(i - 1, -1, -1):
                    recj = log[j]
                    fj = cast(int | float, recj[2])
                    fsj = cast(np.ndarray, recj[3])
                    if (fj <= fi) and (domination(fsi, fsj) > 0):
                        del log[i]
                        break

        header: list[str] = [PROGRESS_FES, PROGRESS_TIME_MILLIS,
                             PROGRESS_CURRENT_F]
        header.extend(f"{PROGRESS_CURRENT_F}{i}" for i in range(
            len(cast(np.ndarray, log[0])[3])))

        with logger.csv(SECTION_PROGRESS, header) as csv:
            for row in log:
                srow = [row[0], _ns_to_ms(cast(int, row[1])
                                          - start_time), row[2]]
                srow.extend([np_to_py_number(n)
                             for n in cast(np.ndarray, row[3])])
                csv.row(srow)

    def __str__(self) -> str:
        return "MOProcessWithoutSearchSpace"



"""
Algorithms and modules based on Frequency Fitness Assignment (FFA).

Frequency Fitness Assignment (FFA) is an algorithm plugin developed by the
team behind `moptipy`. FFA is suitable only for problems where the number
of different objective values is not too high. On such problems, which are
often discrete or combinatorial, it often can improve the result quality
that the optimization algorithms can reach, at the expense of requiring a
larger runtime.

1. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
2. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*. 2022.
   Early Access. https://dx.doi.org/10.1109/TEVC.2022.3191698
3. Thomas Weise, Mingxu Wan, Ke Tang, Pu Wang, Alexandre Devert, and Xin
   Yao. Frequency Fitness Assignment. *IEEE Transactions on Evolutionary
   Computation (IEEE-EC),* 18(2):226-243, April 2014.
   https://dx.doi.org/10.1109/TEVC.2013.2251885
4. Thomas Weise, Yan Chen, Xinlu Li, and Zhize Wu. Selecting a diverse set of
   benchmark instances from a tunable model problem for black-box discrete
   optimization algorithms. *Applied Soft Computing Journal (ASOC),*
   92:106269, June 2020. https://dx.doi.org/10.1016/j.asoc.2020.106269
5. Thomas Weise, Xinlu Li, Yan Chen, and Zhize Wu. Solving Job Shop Scheduling
   Problems Without Using a Bias for Good Solutions. In *Genetic and
   Evolutionary Computation Conference Companion (GECCO'21 Companion),*
   July 10-14, 2021, Lille, France. ACM, New York, NY, USA.
   ISBN 978-1-4503-8351-6. https://dx.doi.org/10.1145/3449726.3463124
6. Thomas Weise, Mingxu Wan, Ke Tang, and Xin Yao. Evolving Exact Integer
   Algorithms with Genetic Programming. In *Proceedings of the IEEE Congress
   on Evolutionary Computation (CEC'14), Proceedings of the 2014 World
   Congress on Computational Intelligence (WCCI'14),* pages 1816-1823,
   July 6-11, 2014, Beijing, China. Los Alamitos, CA, USA: IEEE Computer
   Society Press. ISBN: 978-1-4799-1488-3.
   https://dx.doi.org/10.1109/CEC.2014.6900292
7. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, Sarah Louise
   Thomson, and Thomas Weise. Addressing the Traveling Salesperson Problem
   with Frequency Fitness Assignment and Hybrid Algorithms. *Soft Computing.*
   2024. https://dx.doi.org/10.1007/s00500-024-09718-8
"""



"""Functions that can be used to test spaces."""
from typing import Any, Callable

# noinspection PyPackageRequirements
import pytest
from pycommons.io.csv import COMMENT_START
from pycommons.types import type_error

from moptipy.api.space import Space, check_space
from moptipy.tests.component import validate_component
from moptipy.utils.logger import SECTION_END, SECTION_START


def validate_space(
        space: Space,
        make_element_valid: Callable[[Any], Any] | None = lambda x: x,
        make_element_invalid: Callable[[Any], Any] | None = None) -> None:
    """
    Check whether an object is a moptipy space.

    :param space: the space to test
    :param make_element_valid: a method that can turn a point from the
        space into a valid point
    :param make_element_invalid: a method can a valid point from the
        space into an invalid one
    :raises ValueError: if `space` is not a valid instance of
        :class:`~moptipy.api.space.Space`
    :raises TypeError: if incorrect types are encountered
    """
    if not isinstance(space, Space):
        raise type_error(space, "space", Space)
    check_space(space)
    validate_component(space)

    if not (hasattr(space, "create") and callable(getattr(space, "create"))):
        raise ValueError("space must have method create.")
    x1 = space.create()
    if x1 is None:
        raise ValueError("Spaces must create() valid objects, "
                         "but returned None.")
    x2 = space.create()
    if x2 is None:
        raise ValueError("Spaces must create() valid objects, "
                         "but returned None.")
    if x2 is x1:
        raise ValueError("The create() method must produce different "
                         "instances when invoked twice, but returned the "
                         "same object.")

    if type(x1) is not type(x2):  # noqa: PIE789
        raise ValueError("The create() method must produce instances of "
                         f"the same type, but got {type(x1)} and {type(x2)}.")

    if not (hasattr(space, "copy")
            and callable(getattr(space, "copy"))):
        raise ValueError("space must have method copy.")
    space.copy(x2, x1)

    if not (hasattr(space, "is_equal")
            and callable(getattr(space, "is_equal"))):
        raise ValueError("space must have method is_equal.")
    if not space.is_equal(x1, x2):
        raise ValueError("space.copy(x1, x2) did not lead to "
                         "space.is_equal(x1, x2).")

    if make_element_valid is None:
        return

    x1 = make_element_valid(x1)

    if not (hasattr(space, "validate")
            and callable(getattr(space, "validate"))):
        raise ValueError("space must have method validate.")
    space.validate(x1)

    if not (hasattr(space, "to_str") and callable(getattr(space, "to_str"))):
        raise ValueError("space must have method to_str.")
    strstr = space.to_str(x1)
    if not isinstance(strstr, str):
        raise type_error(strstr, f"space.to_str(x) for {x1}", str)
    if len(strstr) <= 0:
        raise ValueError(
            "space.to_str(x) must not produce empty strings, "
            f"but we got {strstr!r}.")
    if strstr.strip() != strstr:
        raise ValueError(
            "space.to_str(x) must not include leading or trailing spaces,"
            f" but we go {strstr!r}.")
    if SECTION_START in strstr:
        raise ValueError(f"space.to_str() must not include "
                         f"{SECTION_START!r}, but is {strstr!r}.")
    if SECTION_END in strstr:
        raise ValueError(f"space.to_str() must not include "
                         f"{SECTION_END!r}, but is {strstr!r}.")
    if COMMENT_START in strstr:
        raise ValueError(f"space.to_str() must not include "
                         f"{COMMENT_START!r}, but is {strstr!r}.")

    if not (hasattr(space, "from_str")
            and callable(getattr(space, "from_str"))):
        raise ValueError("space must have method from_str.")
    x3 = space.from_str(strstr)
    if (x3 is x1) or (x3 is x2):
        raise ValueError("from_str() cannot return the same object as "
                         "create().")
    if not space.is_equal(x1, x3):
        raise ValueError("from_str(to_str()) must return equal object.")
    if space.to_str(x3) != strstr:
        raise ValueError("to_str(from_str(to_str())) must return same "
                         "string.")
    space.validate(x3)

    if make_element_invalid is None:
        return
    x2 = make_element_invalid(x3)
    if space.is_equal(x1, x2):
        raise ValueError(
            "make_element_invalid did not lead to a change in element!")

    with pytest.raises(ValueError):  # noqa
        space.validate(x2)



"""
A nullary operator shuffling a signed permutation and flipping signs.

This operator first copies the canonical permutation to the destination
string. It then applies the shuffle procedure of the random number
generator, which probably internally applies a Fisher-Yates Shuffle.
The result is a random permutation. Then, it goes over the permutation
once more and the signs of the elements randomly.
"""
from typing import Final

import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.operators import Op0
from moptipy.spaces.signed_permutations import SignedPermutations


class Op0ShuffleAndFlip(Op0):
    """Shuffle permutations randomly and flip signs randomly."""

    def __init__(self, space: SignedPermutations):
        """
        Initialize this shuffle and flip operation: use blueprint from space.

        :param space: the search space
        """
        super().__init__()
        if not isinstance(space, SignedPermutations):
            raise type_error(space, "space", SignedPermutations)
        #: the internal blueprint for filling permutations
        self.__blueprint: Final[np.ndarray] = space.blueprint

    def op0(self, random: Generator, dest: np.ndarray) -> None:
        """
        Copy the base string to `dest` and shuffle it and flip signs randomly.

        :param random: the random number generator
        :param dest: the signed permutation that should be filled with a
            random sequence of the base permutation with potentially flipped
            signs.
        """
        np.copyto(dest, self.__blueprint)
        random.shuffle(dest)  # Shuffle destination array randomly.
        # integers(0, 2, n) -> n values V in {0, 1}
        # (2 * V) - 1 -> a value in {-1, 1}
        # dest[1] * V -> a value which is either {-dest[i], dest[i]}
        dest *= ((2 * random.integers(low=0, high=2, size=len(dest))) - 1)

    def __str__(self) -> str:
        """
        Get the name of this shuffle-and-flip operator.

        :return: "shuffleAndFlip"
        """
        return "shuffleAndFlip"



"""
The Frequency Fitness Assignment (FFA) Process.

Frequency Fitness Assignment (FFA) replaces all objective values with their
encounter frequencies in the selection decisions. The more often an
objective value is encountered, the higher gets its encounter frequency.
Therefore, local optima are slowly receiving worse and worse fitness.

Notice that this implementation of FFA has a slight twist to it:
It will incorporate the iteration index
(:attr:`~moptipy.algorithms.so.record.Record.it`) of the solutions
into the fitness.
This index is used to break ties, in which case newer solutions are preferred.

This can make the EA with FFA compatible with the
:class:`moptipy.algorithms.so.ffa.fea1plus1.FEA1plus1` if "best" selection
(:class:`moptipy.algorithms.modules.selections.best.Best`) is used
at mu=lambda=1.
To facilitate this, there is one special case in the FFA fitness assignment:
If the population consists of exactly 1 element at iteration index 0, then
the frequency values are not updated.

1. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020.
   https://dx.doi.org/10.1109/TEVC.2020.3032090
2. Thomas Weise, Zhize Wu, Xinlu Li, Yan Chen, and Jörg Lässig. Frequency
   Fitness Assignment: Optimization without Bias for Good Solutions can be
   Efficient. *IEEE Transactions on Evolutionary Computation (TEVC)*. 2022.
   Early Access. https://dx.doi.org/10.1109/TEVC.2022.3191698
3. Thomas Weise, Mingxu Wan, Ke Tang, Pu Wang, Alexandre Devert, and Xin
   Yao. Frequency Fitness Assignment. *IEEE Transactions on Evolutionary
   Computation (IEEE-EC),* 18(2):226-243, April 2014.
   https://dx.doi.org/10.1109/TEVC.2013.2251885
4. Thomas Weise, Yan Chen, Xinlu Li, and Zhize Wu. Selecting a diverse set of
   benchmark instances from a tunable model problem for black-box discrete
   optimization algorithms. *Applied Soft Computing Journal (ASOC),*
   92:106269, June 2020. https://dx.doi.org/10.1016/j.asoc.2020.106269
5. Thomas Weise, Xinlu Li, Yan Chen, and Zhize Wu. Solving Job Shop Scheduling
   Problems Without Using a Bias for Good Solutions. In *Genetic and
   Evolutionary Computation Conference Companion (GECCO'21 Companion),*
   July 10-14, 2021, Lille, France. ACM, New York, NY, USA.
   ISBN 978-1-4503-8351-6. https://dx.doi.org/10.1145/3449726.3463124
6. Thomas Weise, Mingxu Wan, Ke Tang, and Xin Yao. Evolving Exact Integer
   Algorithms with Genetic Programming. In *Proceedings of the IEEE Congress
   on Evolutionary Computation (CEC'14), Proceedings of the 2014 World
   Congress on Computational Intelligence (WCCI'14),* pages 1816-1823,
   July 6-11, 2014, Beijing, China. Los Alamitos, CA, USA: IEEE Computer
   Society Press. ISBN: 978-1-4799-1488-3.
   https://dx.doi.org/10.1109/CEC.2014.6900292
"""
from typing import Final

from numpy.random import Generator
from pycommons.types import type_error, type_name_of

from moptipy.algorithms.so.ffa.ffa_h import clear_h, create_h, log_h
from moptipy.algorithms.so.fitness import Fitness, FRecord
from moptipy.api.objective import Objective, check_objective
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection

#: The lower bound at which we switch to an offset-based backing array for
#: the frequency table H.
SWITCH_TO_OFFSET_LB: Final[int] = 8_388_608


class FFA(Fitness):
    """The frequency fitness assignment (FFA) process."""

    def __init__(self, f: Objective, log_h_tbl: bool = False) -> None:
        """
        Create the frequency fitness assignment mapping.

        :param f: the objective function
        """
        super().__init__()
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        check_objective(f)
        h, ofs = create_h(f)
        #: the internal H-table
        self.__h: Final = h
        #: the offset
        self.__ofs: Final[int] = ofs
        #: should we log the H-table?
        self.__log_h_tbl = log_h_tbl
        #: are we in the very first iteration?
        self.__first: bool = True

    def __str__(self):
        """
        Get the name (`"ffa"`) of the FFA fitness assignment process.

        :return: the name of this process: `ffa`
        retval "ffa": always
        """
        return "ffa"

    def log_information_after_run(self, process: Process) -> None:
        """Write the H table."""
        if self.__log_h_tbl:
            log_h(process, self.__h, self.__ofs)

    def assign_fitness(self, p: list[FRecord], random: Generator) -> None:
        """
        Assign the frequency fitness.

        :param p: the list of records
        :param random: ignored

        >>> from moptipy.examples.bitstrings.onemax import OneMax
        >>> fit = FFA(OneMax(200))
        >>> a = FRecord(None, 1)
        >>> b = FRecord(None, 2)
        >>> c = FRecord(None, 2)
        >>> d = FRecord(None, 3)
        >>> from numpy.random import default_rng
        >>> rand = default_rng()
        >>> fit.assign_fitness([a, b, c, d], rand)
        >>> assert a.fitness == 1
        >>> assert b.fitness == 2
        >>> assert c.fitness == 2
        >>> assert d.fitness == 1
        >>> fit.assign_fitness([a, b, c, d], rand)
        >>> assert a.fitness == 2
        >>> assert b.fitness == 4
        >>> assert c.fitness == 4
        >>> assert d.fitness == 2
        """
        h: Final = self.__h
        ofs: Final[int] = self.__ofs

        min_it = max_it = p[0].it  # the minimum and maximum iteration index
        for r in p:
            it: int = r.it  # get iteration index from record
            if it < min_it:  # update minimum iteration index
                min_it = it
            elif it > max_it:  # update maximum iteration index
                max_it = it
# Compatibility with (1+1) FEA: In first generation with a population size of
# 1, all elements get the same fitness zero. Otherwise, we would count the
# very first solution in the next fitness assignment step again.
# This creates a small incompatibility between FFA with mu=1 and FFA with m>1.
# This incompatibility is not nice, but it is the only way to ensure that the
# (1+1) FEA and the GeneralEA with EA and mu=1, lambda=1 are identical.
        first = self.__first
        self.__first = False
        if first and (max_it <= 0) and (len(p) <= 1):
            for r in p:  # same fitness: 0
                r.fitness = 0
            return
        for r in p:
            h[r.f + ofs] += 1  # type: ignore
        it_range: Final[int] = max_it - min_it + 1  # range of it index
        for r in p:
            r.fitness = ((int(h[r.f + ofs]) * it_range)  # type: ignore
                         + max_it - r.it)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("ofs", self.__ofs)
        logger.key_value("hType", type_name_of(self.__h))

    def initialize(self) -> None:
        """Initialize the algorithm."""
        super().initialize()
        clear_h(self.__h)
        self.__first = True



"""
The `1rs` "algorithm" creates one single random solution.

The single random sample algorithm applies the nullary search operator,
an implementation of :meth:`~moptipy.api.operators.Op0.op0`,
to sample exactly one single random solution. It then evaluates the
solution by passing it to
:meth:`~moptipy.api.process.Process.evaluate`. It then terminates.

This is a very very bad optimization algorithm. We only use it in our
book to illustrate one basic concept for solving optimization problems:
The generation of random solutions. The single-random sampling
algorithm here is actually very wasteful: since it only generates
exactly one single solution, it does not use its computational budget
well. Even if you grant it 10'000 years, it will still only generate
one solution. Even if it could generate and test thousands or millions
of solutions, it will not do it. Nevertheless, after applying this
"algorithm," you will have one valid solution remembered in the
optimization process (embodied as instance `process` of
:class:`~moptipy.api.process.Process`).

This concept of random sampling is then refined in the
:class:`~moptipy.algorithms.random_sampling.RandomSampling` as the `rs`
algorithm, which repeats generating random solutions until its allotted
runtime is exhausted.

1. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
"""
from typing import Final

from moptipy.api.algorithm import Algorithm0
from moptipy.api.operators import Op0
from moptipy.api.process import Process


# start book
class SingleRandomSample(Algorithm0):
    """An algorithm that creates one single random solution."""

    def solve(self, process: Process) -> None:
        """
        Apply single random sampling to an optimization problem.

        :param process: the black-box process object
        """
        x: Final = process.create()  # Create the solution record.
        self.op0.op0(process.get_random(), x)  # Create random solution
        process.evaluate(x)  # Evaluate that random solution.
# end book

    def __init__(self, op0: Op0) -> None:
        """
        Create the single random sample algorithm.

        :param op0: the nullary search operator
        """
        super().__init__("1rs", op0)



"""
An operator deleting an element from a permutation and inserting it elsewhere.

The operator is similar to a combination of the `rol` and `ror` operators
in [1]. If the permutation represents, e.g., a tour in the Traveling
Salesperson Problem (TSP), then a rotation replaces two edges (if two
neighboring elements are rotated = swapped) or three edges (if at least three
elements take part in the rotation). It may therefore be considered as a
possible 3-opt move on the TSP.

An alternative way to look at this is given in [2, 3, 4], where the operator
is called "Insertion Mutation": An element of the permutation is removed from
its current position (`i1`) and inserted at another position (`i2`). The
operator is therefore called "position based mutation operator" in [5].

1. Thomas Weise, Raymond Chiong, Ke Tang, Jörg Lässig, Shigeyoshi Tsutsui,
   Wenxiang Chen, Zbigniew Michalewicz, and Xin Yao. Benchmarking Optimization
   Algorithms: An Open Source Framework for the Traveling Salesman Problem.
   *IEEE Computational Intelligence Magazine (CIM)* 9(3):40-52, August 2014.
   https://doi.org/10.1109/MCI.2014.2326101
2. Pedro Larrañaga, Cindy M. H. Kuijpers, Roberto H. Murga, I. Inza, and
   S. Dizdarevic. Genetic Algorithms for the Travelling Salesman Problem: A
   Review of Representations and Operators. *Artificial Intelligence Review,*
   13(2):129-170, April 1999. Kluwer Academic Publishers, The Netherlands.
   https://doi.org/10.1023/A:1006529012972
3. David B. Fogel. An Evolutionary Approach to the Traveling Salesman Problem.
   *Biological Cybernetics* 60(2):139-144, December 1988.
   https://doi.org/10.1007/BF00202901
4. Zbigniew Michalewicz. *Genetic Algorithms + Data Structures = Evolution
   Programs,* Berlin, Germany: Springer-Verlag GmbH. 1996. ISBN:3-540-58090-5.
   https://doi.org/10.1007/978-3-662-03315-9
5. Gilbert Syswerda. Schedule Optimization Using Genetic Algorithms. In
   Lawrence Davis, (ed.), *Handbook of Genetic Algorithms,* pages 332-349.
   1991. New York, NY, USA: Van Nostrand Reinhold.
6. Yuezhong Wu, Thomas Weise, and Raymond Chiong. Local Search for the
   Traveling Salesman Problem: A Comparative Study. In *Proceedings of the
   14th IEEE Conference on Cognitive Informatics & Cognitive Computing
   (ICCI*CC'15),* July 6-8, 2015, Beijing, China, pages 213-220, Los Alamitos,
   CA, USA: IEEE Computer Society Press, ISBN: 978-1-4673-7289-3.
   https://doi.org/10.1109/ICCI-CC.2015.7259388
7. Dan Xu, Thomas Weise, Yuezhong Wu, Jörg Lässig, and Raymond Chiong. An
   Investigation of Hybrid Tabu Search for the Traveling Salesman Problem. In
   *Proceedings of the 10th International Conference on Bio-Inspired Computing
   &mdash; Theories and Applications (BIC-TA'15),* September 25-28, 2015,
   Hefei, Anhui, China, volume 562 of Communications in Computer and
   Information Science. Berlin/Heidelberg: Springer-Verlag, pages 523-537,
   ISBN 978-3-662-49013-6. https://doi.org/10.1007/978-3-662-49014-3_47
8. Weichen Liu, Thomas Weise, Yuezhong Wu, and Raymond Chiong. Hybrid Ejection
   Chain Methods for the Traveling Salesman Problem. In *Proceedings of the
   10th International Conference on Bio-Inspired Computing &mdash; Theories
   and Applications (BIC-TA'15),* September 25-28, 2015, Hefei, Anhui, China,
   volume 562 of Communications in Computer and Information Science.
   Berlin/Heidelberg: Springer-Verlag, pages 268-282, ISBN 978-3-662-49013-6.
   https://doi.org/10.1007/978-3-662-49014-3_25
9. Yuezhong Wu, Thomas Weise, and Weichen Liu. Hybridizing Different Local
   Search Algorithms with Each Other and Evolutionary Computation: Better
   Performance on the Traveling Salesman Problem. In *Proceedings of the 18th
   Genetic and Evolutionary Computation Conference (GECCO'16),* July 20-24,
   2016, Denver, CO, USA, pages 57-58, New York, NY, USA: ACM.
   ISBN: 978-1-4503-4323-7. https://doi.org/10.1145/2908961.2909001
"""
from typing import Callable, Final

import numba  # type: ignore
import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def try_single_rotate(arr: np.ndarray, i1: int, i2: int) -> bool:  # +book
    """
    Rotate a portion of an array to the left or right in place.

    If `i1 < i2`, then a left rotation by one step is performed. In other
    words, the element at index `i1 + 1` goes to index `i1`, the element at
    index `i1 + 2` goes to index `i1 + 1`, and so on. The lst element, i.e.,
    the one at index `i2` goes to index `i2 - 1`. Finally, the element that
    originally was at index `i1` goes to index `i2`. If any element in the
    array has changed, this function returns `False`, otherwise `True`.

    If `i1 > i2`, then a right rotation by one step is performed. In other
    words, the element at index `i1 - 1` goes to index `i1`, the element at
    index `i1 - 2` goes to index `i1 - 1`, and so on. Finally, the element
    that originally was at index `i1` goes to index `i2`. If any element in
    the array has changed, this function returns `False`, otherwise `True`.

    This corresponds to extracting the element at index `i1` and re-inserting
    it at index `i2`.

    :param arr: the array to rotate
    :param i1: the start index, in `0..len(arr)-1`
    :param i2: the end index, in `0..len(arr)-1`
    :returns: whether the array was *unchanged*
    :retval False: if the array `arr` is now different from before
    :retval True: if the array `arr` has not changed

    >>> import numpy as npx
    >>> dest = npx.array(range(10))
    >>> print(dest)
    [0 1 2 3 4 5 6 7 8 9]
    >>> try_single_rotate(dest, 3, 4)
    False
    >>> print(dest)
    [0 1 2 4 3 5 6 7 8 9]
    >>> try_single_rotate(dest, 3, 4)
    False
    >>> print(dest)
    [0 1 2 3 4 5 6 7 8 9]
    >>> try_single_rotate(dest, 4, 3)
    False
    >>> print(dest)
    [0 1 2 4 3 5 6 7 8 9]
    >>> try_single_rotate(dest, 4, 3)
    False
    >>> print(dest)
    [0 1 2 3 4 5 6 7 8 9]
    >>> try_single_rotate(dest, 3, 6)
    False
    >>> print(dest)
    [0 1 2 4 5 6 3 7 8 9]
    >>> try_single_rotate(dest, 6, 3)
    False
    >>> print(dest)
    [0 1 2 3 4 5 6 7 8 9]
    >>> try_single_rotate(dest, 0, len(dest) - 1)
    False
    >>> print(dest)
    [1 2 3 4 5 6 7 8 9 0]
    >>> try_single_rotate(dest, len(dest) - 1, 0)
    False
    >>> print(dest)
    [0 1 2 3 4 5 6 7 8 9]
    >>> try_single_rotate(dest, 7, 7)
    True
    >>> dest = np.array([0, 1, 2, 3, 3, 3, 3, 3, 8, 9])
    >>> try_single_rotate(dest, 7, 7)
    True
    >>> try_single_rotate(dest, 4, 6)
    True
    >>> print(dest)
    [0 1 2 3 3 3 3 3 8 9]
    >>> try_single_rotate(dest, 6, 4)
    True
    >>> print(dest)
    [0 1 2 3 3 3 3 3 8 9]
    >>> try_single_rotate(dest, 4, 7)
    True
    >>> print(dest)
    [0 1 2 3 3 3 3 3 8 9]
    >>> try_single_rotate(dest, 6, 7)
    True
    >>> print(dest)
    [0 1 2 3 3 3 3 3 8 9]
    >>> try_single_rotate(dest, 4, 8)
    False
    >>> print(dest)
    [0 1 2 3 3 3 3 8 3 9]
    >>> try_single_rotate(dest, 8, 4)
    False
    >>> print(dest)
    [0 1 2 3 3 3 3 3 8 9]
    >>> try_single_rotate(dest, 9, 4)
    False
    >>> print(dest)
    [0 1 2 3 9 3 3 3 3 8]
    >>> try_single_rotate(dest, 4, 9)
    False
    >>> print(dest)
    [0 1 2 3 3 3 3 3 8 9]
    """
    # start book
    if i1 == i2:  # nothing to be done
        return True  # array will not be changed

    unchanged: bool = True  # initially, assume that there is no change

    if i1 < i2:  # rotate to the left: move elements to lower indices?
        first = arr[i1]  # get the element to be removed
        while i1 < i2:  # iterate the indices
            i3 = i1 + 1  # get next higher index
            cpy = arr[i3]  # get next element at that higher index
            unchanged = unchanged and (cpy == arr[i1])  # is a change?
            arr[i1] = cpy  # store next element at the lower index
            i1 = i3  # move to next higher index
        unchanged = unchanged and (first == arr[i2])  # check if change
        arr[i2] = first  # store removed element at highest index
        return unchanged  # return True if something changed, else False

    last = arr[i1]  # last element; rotate right: move elements up
    while i2 < i1:  # iterate over indices
        i3 = i1 - 1  # get next lower index
        cpy = arr[i3]  # get element at that lower index
        unchanged = unchanged and (cpy == arr[i1])  # is a change?
        arr[i1] = cpy  # store element at higher index
        i1 = i3  # move to next lower index
    unchanged = unchanged and (last == arr[i2])  # check if change
    arr[i2] = last  # store removed element at lowest index
    return unchanged  # return True if something changed, else False
# end book


# Temporary fix for https://github.com/numba/numba/issues/9103
def rotate(random: Generator, dest: np.ndarray,  # +book
           x: np.ndarray) -> None:  # +book
    """
    Copy `x` into `dest` and then rotate a subsequence by one step.

    :param random: the random number generator
    :param dest: the array to receive the modified copy of `x`
    :param x: the existing point in the search space

    >>> rand = np.random.default_rng(10)
    >>> xx = np.array(range(10), int)
    >>> out = np.empty(len(xx), int)
    >>> rotate(rand, out, xx)
    >>> print(out)
    [0 1 2 3 4 5 6 8 9 7]
    >>> rotate(rand, out, xx)
    >>> print(out)
    [0 1 2 3 4 5 6 8 7 9]
    >>> rotate(rand, out, xx)
    >>> print(out)
    [0 5 1 2 3 4 6 7 8 9]
    >>> rotate(rand, out, xx)
    >>> print(out)
    [0 1 2 3 4 8 5 6 7 9]
    """
    # start book
    dest[:] = x[:]
    length: Final[int] = len(dest)  # Get the length of `dest`.
    rint: Callable[[int, int], int] = random.integers  # fast call

    # try to rotate the dest array until something changes
    while try_single_rotate(dest, rint(0, length), rint(0, length)):
        pass  # do nothing in the loop, but try rotating again


class Op1Insert1(Op1):
    """
    Delete an element from a permutation and insert it elsewhere.

    This operation keep drawing to random numbers, `i1` and `i2`. If
    `i1 < i2`, then the permutation is rotated one step to the left between
    `i1` and `i2`. If `i1 > i2`, then it is rotated one step to the right
    between `i1` and `i2`. If the permutation has changed, the process is
    stopped. This corresponds to extracting the element at index `i1` and
    re-inserting it at index `i2`. This operator also works for permutations
    with repetitions.
    """

    def __init__(self) -> None:
        """Initialize the object."""
        super().__init__()
        self.op1 = rotate  # type: ignore  # use function directly
# end book

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :returns: "rot1", the name of this operator
        :retval "rot1": always
        """
        return "rot1"



"""
Load the encounter frequencies or the set of different objective values.

This tool can load the different objective values that exist or are
encountered by optimization processes. This may be useful for statistical
evaluations or fitness landscape analyses.

This tool is based on code developed by Mr. Tianyu LIANG (梁天宇),
MSc student at the Institute of Applied Optimization (IAO,
应用优化研究所) of the School of Artificial Intelligence and
Big Data (人工智能与大数据学院) of Hefei University (合肥学院).
"""
import argparse
from collections import Counter
from gc import collect
from math import isfinite
from typing import Any, Callable, Final, Iterable

from pycommons.io.console import logger
from pycommons.io.csv import CSV_SEPARATOR, SCOPE_SEPARATOR
from pycommons.io.path import Path
from pycommons.strings.string_conv import str_to_num
from pycommons.types import type_error

from moptipy.algorithms.so.ffa.ffa_h import H_LOG_SECTION
from moptipy.api.logging import (
    KEY_F_LOWER_BOUND,
    KEY_F_UPPER_BOUND,
    KEY_INSTANCE,
    PROGRESS_CURRENT_F,
    SCOPE_OBJECTIVE_FUNCTION,
    SECTION_PROGRESS,
)
from moptipy.evaluation.base import (
    MultiRunData,
    PerRunData,
)
from moptipy.evaluation.log_parser import SetupAndStateParser
from moptipy.utils.help import moptipy_argparser

#: the lower bound of the objective function
_FULL_KEY_LOWER_BOUND: Final[str] = \
    f"{SCOPE_OBJECTIVE_FUNCTION}{SCOPE_SEPARATOR}{KEY_F_LOWER_BOUND}"
#: the upper bound of the objective function
_FULL_KEY_UPPER_BOUND: Final[str] = \
    f"{SCOPE_OBJECTIVE_FUNCTION}{SCOPE_SEPARATOR}{KEY_F_UPPER_BOUND}"


def from_logs(
        path: str, consumer: Callable[[PerRunData, Counter[
            int | float]], Any],
        report_progress: bool = True,
        report_lower_bound: bool = False,
        report_upper_bound: bool = False,
        report_goal_f: bool = False,
        report_h: bool = True,
        per_instance_known: Callable[[
            str], Iterable[int | float]] = lambda _: ()) -> None:
    """
    Parse a path, pass all discovered objective values per-run to a consumer.

    This function parses the log files in a directory recursively. For each
    log file, it produces a `Counter` filled with all encountered objective
    values and their "pseudo" encounter frequencies. "pseudo" because the
    values returned depend very much on how the function is configured.

    First, if all other parameters are set to `False`, the function passes
    a `Counter` to the `consumer` where the best encountered objective value
    has frequency `1` and no other data is present.
    If `report_progress` is `True`, then each time any objective value is
    encountered in the `PROGRESS` section, its counter is incremented by `1`.
    If the `PROGRESS` section is present, that it. The best encountered
    objective value will have a count of at least one either way.
    If `report_goal_f`, `report_lower_bound`, or `report_upper_bound` are
    `True`, then it is ensured that the goal objective value of the
    optimization process, the lower bound of the objective function, or
    the upper bound of the objective function will have a corresponding count
    of at least `1` if they are present in the log files (in the `SETUP`
    section).
    If `report_h` is `True`, then a frequency fitness assignment `H` section
    is parsed, if present (see :mod:`~moptipy.algorithms.so.ffa.fea1plus1`).
    Such a section contains tuples of objective values and encounter
    frequencies.
    These encounter frequencies are added to the counter. This means that if
    you set *both* `report_progress` and `report_h` to `True`, you will get
    frequencies that are too high.
    Finally, the function `per_instance_known` may return a set of known
    objective values for a given instance (based on its parameter, the
    instance name). Each such objective value will have a frequency of at
    least `1`.

    Generally, if we want the actual encounter frequencies of objective
    values, we could log *all FEs* to the log files and set `report_progress`
    to `True` and everything else to `False`. Then we get correct encounter
    frequencies. Alternatively, if we have a purly FFA-based algorithm (see,
    again, :mod:`~moptipy.algorithms.so.ffa.fea1plus1`), then we can set
    `report_progress` to `True` and everything else to `False` to get a
    similar result, but the encounter frequencies then depend on the selection
    scheme. Alternatively, if we only care about whether an objective value
    was encountered or not, we can simply set both to `True`. Finally, if we
    want to get all *possible* objective values, then we may also set
    `report_goal_f`, `report_lower_bound`, or `report_upper_bound` to `True`
    **if** we are sure that the corresponding objective values do actually
    exist (and are not just bounds that can never be reached).

    :param path: the path to parse
    :param consumer: the consumer receiving, for each log file, an instance of
        :class:`~moptipy.evaluation.base.PerRunData` identifying the run and
        a dictionary with the objective values and lower bounds of their
        existence or encounter frequency. Warning: The dictionary will be
        cleared and re-used for all files.
    :param report_progress: should all values in the `PROGRESS` section be
        reported, if such section exists?
    :param report_lower_bound: should the lower bound reported, if any lower
        bound for the objective function is listed?
    :param report_upper_bound: should the upper bound reported, if any upper
        bound for the objective function is listed?
    :param report_h: should all values in the `H` section be reported, if such
        section exists?
    :param report_goal_f: should we report the goal objective value, if it is
        specified?
    :param per_instance_known: a function that returns a set of known
        objective values per instance
    """
    __InnerLogParser(consumer, report_progress, report_lower_bound,
                     report_upper_bound, report_h, report_goal_f,
                     per_instance_known).parse(path)


def aggregate_from_logs(
        path: str, consumer: Callable[[MultiRunData, Counter[
            int | float]], Any],
        per_instance: bool = True,
        per_algorithm_instance: bool = True,
        report_progress: bool = True,
        report_lower_bound: bool = False,
        report_upper_bound: bool = False,
        report_goal_f: bool = False,
        report_h: bool = True,
        per_instance_known: Callable[[
            str], Iterable[int | float]] = lambda _: ()) -> None:
    """
    Parse a path, aggregate all discovered objective values to a consumer.

    A version of :func:`from_logs` that aggregates results on a per-instance
    and/or per-algorithm-instance combination. The basic process of loading
    the data is described in :func:`from_logs`.

    :param path: the path to parse
    :param consumer: the consumer receiving the aggregated results
    :param per_instance: pass results to the consumer that are aggregated over
        all algorithms and setups and runs for a given instance
    :param per_algorithm_instance: pass results to the consumer that are
        aggregated over all runs and setups for a given algorithm-instance
        combination
    :param report_progress: see :func:`from_logs`
    :param report_lower_bound: see :func:`from_logs`
    :param report_upper_bound: see :func:`from_logs`
    :param report_h: see :func:`from_logs`
    :param report_goal_f: see :func:`from_logs`
    :param per_instance_known: see :func:`from_logs`
    """
    if not callable(consumer):
        raise type_error(consumer, "consumer", call=True)
    if not isinstance(per_instance, bool):
        raise type_error(per_instance, "per_instance", bool)
    if not isinstance(per_algorithm_instance, bool):
        raise type_error(
            per_algorithm_instance, "per_algorithm_instance", bool)

    collection: Final[dict[tuple[str, int, str], Counter[int | float]]] = {}
    runs: Final[Counter[tuple[str, int, str]]] = Counter()

    def __consume(d: PerRunData, c: Counter[int | float]) -> None:
        nonlocal collection
        nonlocal runs
        nonlocal per_instance
        nonlocal per_algorithm_instance

        inst: Final[str] = d.instance
        ai: tuple[str, int, str]
        if per_instance:
            ai = (inst, 1, "")
            runs[ai] += 1
            if ai in collection:
                collection[ai] += c
            else:
                collection[ai] = Counter(c)
        if per_algorithm_instance:
            ai = (inst, 0, d.algorithm)
            runs[ai] += 1
            if ai in collection:
                collection[ai] += c
            else:
                collection[ai] = Counter(c)

    from_logs(path, __consume, report_progress, report_lower_bound,
              report_upper_bound, report_goal_f, report_h, per_instance_known)

    for key in sorted(collection.keys()):
        val = collection[key]
        del collection[key]
        consumer(MultiRunData(key[2] if key[2] else None,
                              key[0], None, None, runs[key]), val)
        del val


def number_of_objective_values_to_csv(
        input_dir: str, output_file: str,
        per_instance: bool = True,
        per_algorithm_instance: bool = True,
        report_lower_bound: bool = False,
        report_upper_bound: bool = False,
        report_goal_f: bool = False,
        per_instance_known: Callable[[
            str], Iterable[int | float]] = lambda _: ()) -> None:
    """
    Print the number of unique objective values to a CSV file.

    A version of :func:`aggregate_from_logs` that collects the existing
    objective values and prints an overview to a file.

    :param input_dir: the path to parse
    :param output_file: the output file to generate
    :param per_instance: pass results to the consumer that are aggregated over
        all algorithms and setups and runs for a given instance
    :param per_algorithm_instance: pass results to the consumer that are
        aggregated over all runs and setups for a given algorithm-instance
        combination
    :param report_lower_bound: see :func:`from_logs`
    :param report_upper_bound: see :func:`from_logs`
    :param report_goal_f: see :func:`from_logs`
    :param per_instance_known: see :func:`from_logs`
    """
    input_path: Final[Path] = Path(input_dir)
    output_path: Final[Path] = Path(output_file)
    logger(f"Collecting number of objective values from {input_path!r} "
           f"in {output_path!r}.")
    logger(f"Lower bounds will{'' if report_lower_bound else ' not'} be "
           "treated as existing objective values.")
    logger(f"Upper bounds will{'' if report_upper_bound else ' not'} be "
           "treated as existing objective values.")
    logger("Goal objective values bounds will"
           f"{'' if report_upper_bound else ' not'} be treated as existing "
           "objective values.")
    output_path.ensure_parent_dir_exists()
    data: Counter[tuple[str, str]] = Counter()
    instances: set[str] = set()
    algorithms: set[str] = set()

    def __collector(d: MultiRunData, c: Counter[int | float]) -> None:
        nonlocal data
        nonlocal instances
        nonlocal algorithms
        inst: str = ""
        if d.instance is not None:
            inst = d.instance
            instances.add(inst)
        algo: str = ""
        if d.algorithm is not None:
            algo = d.algorithm
            algorithms.add(algo)
        data[(inst, algo)] += len(c)

    aggregate_from_logs(
        input_path, __collector, per_instance, per_algorithm_instance,
        True, report_lower_bound, report_upper_bound, report_goal_f, True,
        per_instance_known)

    algos: Final[list[str]] = sorted(algorithms)

    logger(f"Now writing information gathered for {len(algos)} "
           f"algorithms and {len(instances)} to {output_path!r}.")
    with output_path.open_for_write() as ow:
        wrt = ow.write

        # write the header
        wrt(KEY_INSTANCE)
        for a in algos:
            wrt(CSV_SEPARATOR)
            wrt(a)
        wrt(CSV_SEPARATOR)
        wrt("all")
        wrt("\n")

        # write the output
        for instance in sorted(instances):
            wrt(instance)
            for a in algos:
                wrt(CSV_SEPARATOR)
                wrt(str(data[(instance, a)]))
            wrt(CSV_SEPARATOR)
            wrt(str(data[(instance, "")]))
            wrt("\n")
    logger(f"Finished writing {output_path!r}.")


class __InnerLogParser(SetupAndStateParser):
    """The internal log parser class for gathering objective values."""

    def __init__(self,
                 consumer: Callable[[PerRunData, Counter[int | float]], Any],
                 report_progress: bool, report_lower_bound: bool,
                 report_upper_bound: bool, report_h: bool,
                 report_goal_f: bool, per_instance_known: Callable[[
                     str], Iterable[int | float]]):
        """
        Create the internal log parser.

        :param consumer: the consumer
        :param report_progress: should all values in the `PROGRESS` section be
            reported, if such section exists?
        :param report_lower_bound: should the lower bound reported, if any
            lower bound for the objective function is listed?
        :param report_upper_bound: should the upper bound reported, if any
            upper bound for the objective function is listed?
        :param report_h: should all values in the `H` section be reported, if
            such section exists?
        :param report_goal_f: should we report the goal objective value, if it
            is specified?
        :param per_instance_known: a function that returns a set of known
            objective values per instance
        """
        super().__init__()
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        if not isinstance(report_progress, bool):
            raise type_error(report_progress, "report_progress", bool)
        if not isinstance(report_lower_bound, bool):
            raise type_error(report_lower_bound, "report_lower_bound", bool)
        if not isinstance(report_upper_bound, bool):
            raise type_error(report_upper_bound, "report_upper_bound", bool)
        if not isinstance(report_h, bool):
            raise type_error(report_h, "report_h", bool)
        if not isinstance(report_goal_f, bool):
            raise type_error(report_goal_f, "report_goal_f", bool)
        if not callable(per_instance_known):
            raise type_error(
                per_instance_known, "per_instance_known", call=True)
        #: the consumer
        self.__consumer: Final[Callable[[PerRunData, Counter], Any]] = consumer
        #: report the progress
        self.__report_progress: Final[bool] = report_progress
        #: report the frequency history
        self.__report_h: Final[bool] = report_h
        #: report the lower bound, if any
        self.__report_lower_bound: Final[bool] = report_lower_bound
        #: report the upper bound, if any
        self.__report_upper_bound: Final[bool] = report_upper_bound
        #: report the goal objective value
        self.__report_goal_f: Final[bool] = report_goal_f
        #: the per-instance known objective values
        self.__per_instance_known: Final[Callable[
            [str], Iterable[int | float]]] = per_instance_known
        #: the internal counter
        self.__counter: Final[Counter[int | float]] = Counter()
        #: the internal state variable
        self.__state_progress: int = 0
        #: the internal state variable
        self.__state_h: int = 0

    def start_file(self, path: Path) -> bool:
        """
        Begin parsing the file identified by `path`.

        :param path: the path identifying the file
        """
        if not super().start_file(path):
            return False
        self.__state_progress = 0 if self.__report_progress else -1
        self.__state_h = 0 if self.__report_h else -1
        return True

    def parse_dir(self, path: str) -> bool:
        collect()
        ret: Final[bool] = super().parse_dir(path)
        collect()
        return ret

    def process(self) -> None:
        counter: Final[Counter[int | float]] = self.__counter

        # report goal objective value, if encountered
        if (self.__report_goal_f and (self.goal_f is not None)
                and isfinite(self.goal_f)):
            counter[self.goal_f] = max(counter[self.goal_f], 1)
        # add the known values
        for val in self.__per_instance_known(self.instance):
            if isfinite(val):
                counter[val] = max(counter[val], 1)
        # add the best objective value
        if (self.best_f is not None) and isfinite(self.best_f):
            counter[self.best_f] = max(counter[self.best_f], 1)

        self.__consumer(PerRunData(
            algorithm=self.algorithm,
            instance=self.instance,
            objective=self.objective,
            encoding=self.encoding,
            rand_seed=self.rand_seed), counter)
        counter.clear()
        collect()

    def start_section(self, title: str) -> bool:
        if title == SECTION_PROGRESS:
            if self.__state_progress >= 2:
                raise ValueError(f"Already did section {title}.")
            if self.__state_progress < 0:
                self.__state_progress = 2
                return False
            self.__state_progress = 1
            return True
        if title == H_LOG_SECTION:
            if self.__state_h >= 2:
                raise ValueError(f"Already did section {title}.")
            if self.__state_h < 0:
                self.__state_h = 2
                return False
            self.__state_h = 1
            return True
        return super().start_section(title)

    def setup_section(self, data: dict[str, str]) -> None:
        """
        Parse the data from the `setup` section.

        :param data: the parsed data
        """
        super().setup_section(data)
        if self.__report_lower_bound and (_FULL_KEY_LOWER_BOUND in data):
            lb: Final[int | float] = str_to_num(
                data[_FULL_KEY_LOWER_BOUND])
            if isfinite(lb):
                self.__counter[lb] = max(self.__counter[lb], 1)
        if self.__report_upper_bound and (_FULL_KEY_UPPER_BOUND in data):
            ub: Final[int | float] = str_to_num(
                data[_FULL_KEY_UPPER_BOUND])
            if isfinite(ub):
                self.__counter[ub] = max(self.__counter[ub], 1)

    def needs_more_lines(self) -> bool:
        return ((self.__state_h == 0) or (self.__state_progress == 0)
                or super().needs_more_lines())

    def lines(self, lines: list[str]) -> bool:
        if not isinstance(lines, list):
            raise type_error(lines, "lines", list)

        counter: Counter[int | float]
        if self.__state_progress == 1:
            self.__state_progress = 2
            columns = [c.strip() for c in lines[0].split(CSV_SEPARATOR)]
            f_col: Final[int] = columns.index(PROGRESS_CURRENT_F)
            counter = self.__counter
            for line in lines[1:]:
                f: int | float = str_to_num(line.split(
                    CSV_SEPARATOR)[f_col])
                counter[f] += 1
        elif self.__state_h == 1:
            self.__state_h = 2
            counter = self.__counter
            for line in lines:
                split = line.split(CSV_SEPARATOR)
                prev_f: int | float = -1
                for i in range(0, len(split), 2):
                    cis = split[i]
                    if str.__len__(cis) <= 0:
                        cur_f = prev_f + 1
                    else:
                        cur_f = str_to_num(cis)
                    prev_f = cur_f
                    counter[cur_f] += int(split[i + 1])
        else:
            return super().lines(lines)

        return self.needs_more_lines()


# Print a CSV file
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipy_argparser(
        __file__, "Collecting the Number of Existing Objective Values.",
        "Gather all the existing objective values and store them in a "
        "CSV-formatted file.")
    parser.add_argument(
        "source", nargs="?", default="./results",
        help="the location of the experimental results, i.e., the root folder "
             "under which to search for log files", type=Path)
    parser.add_argument(
        "dest", help="the path to the end results CSV file to be created",
        type=Path, nargs="?",
        default="./evaluation/objective_values.txt")
    parser.add_argument(
        "--lb", help="count the lower bound of the "
                     "objective as objective value",
        action="store_true")
    parser.add_argument(
        "--ub", help="count the upper bound of the "
                     "objective as objective value",
        action="store_true")
    parser.add_argument(
        "--goal", help="count the goal objective value as "
                       "existing objective value",
        action="store_true")
    args: Final[argparse.Namespace] = parser.parse_args()
    number_of_objective_values_to_csv(
        args.source, args.dest, per_instance=True,
        per_algorithm_instance=True, report_lower_bound=args.lb,
        report_upper_bound=args.ub, report_goal_f=args.goal)



"""
Components for parsing and evaluating log files generated by experiments.

Via the :mod:`moptipy.api`, it is possible to log the progress or end
results of optimization algorithms runs in text-based log files.
With the methods in this package here, you can load and evaluate them. This
usually follows a multi-step approach: For example, you can first extract the
end results from several algorithms and instances into a single file via the
:class:`~moptipy.evaluation.end_results.EndResult`. This could then be
processed to per-algorithm or per-instance statistics using
:class:`~moptipy.evaluation.end_statistics.EndStatistics`.
"""


def __lang_setup() -> None:
    """Perform the internal language setup."""
    # pylint: disable=C0415
    import pycommons.math.sample_statistics as ss  # noqa

    import moptipy.api.logging as lg  # pylint: disable=C0415  # noqa
    import moptipy.evaluation.base as bs  # pylint: disable=C0415  # noqa
    import moptipy.evaluation.end_statistics as es  # pylint: disable=C0415  # noqa
    from moptipy.utils.lang import DE, EN, ZH  # pylint: disable=C0415  # noqa

    EN.extend({
        "f": "f",
        "bestF": "f",
        "feasible": "feasible",
        "name": "name",
        "time": "time",
        "time_in_fes": "time in FEs",
        "time_in_ms": "time in ms",
        "lower_bound": "lower bound",
        "lower_bound_short": "lb",
        bs.F_NAME_NORMALIZED: "normalized f",
        bs.F_NAME_SCALED: "scaled f",
        bs.F_NAME_RAW: "f",
        bs.TIME_UNIT_FES: "time in FEs",
        bs.TIME_UNIT_MILLIS: "time in ms",
        lg.KEY_TOTAL_TIME_MILLIS: "total time in ms",
        lg.KEY_LAST_IMPROVEMENT_TIME_MILLIS: "last improvement time at ms",
        lg.KEY_TOTAL_FES: "total time in FEs",
        lg.KEY_LAST_IMPROVEMENT_FE: "last improvement at FE",
        "algorithm_on_instance": "algorithm \u00d7 instance",
        "ERT": "ERT",
        es.KEY_ERT_FES: "ERT[FEs]",
        es.KEY_ERT_TIME_MILLIS: "ERT[ms]",
        "ECDF": "ECDF",
        "setup": "setup",
        "best": "best",
        "worst": "worst",
        "summary": "summary",
        "all_insts": "all instances",
        "all_algos": "all algorithms",
        "single_run": "single run",
        ss.KEY_MEAN_GEOM: "gmean",
        ss.KEY_STDDEV: "sd",
        ss.KEY_MEDIAN: "median",
        ss.KEY_MAXIMUM: "max",
        ss.KEY_MINIMUM: "min",
        ss.KEY_MEAN_ARITH: "mean",
        es.KEY_N_SUCCESS: "#successes",
    })

    DE.extend({
        "f": "f",
        "bestF": "f",
        "feasible": "realisierbar",
        "name": "Name",
        "time": "Zeit",
        "time_in_ms": "Zeit in ms",
        "time_in_fes": "Zeit in FEs",
        "lower_bound": "untere Schranke",
        "lower_bound_short": "us",
        bs.F_NAME_NORMALIZED: "normalisierte f",
        bs.F_NAME_SCALED: "skalierte f",
        bs.F_NAME_RAW: "f",
        bs.TIME_UNIT_FES: "Zeit in FEs",
        bs.TIME_UNIT_MILLIS: "Zeit in ms",
        lg.KEY_TOTAL_TIME_MILLIS: "Gesamtzeit in ms",
        lg.KEY_LAST_IMPROVEMENT_TIME_MILLIS: "letzte Verbesserung bei ms",
        lg.KEY_TOTAL_FES: "Gesamtzeit in FEs",
        lg.KEY_LAST_IMPROVEMENT_FE: "letzte Verbesserung bei FE",
        "algorithm_on_instance": "Algorithmus \u00d7 Instanz",
        "ERT": "ERT",
        es.KEY_ERT_FES: "ERT[FEs]",
        es.KEY_ERT_TIME_MILLIS: "ERT[ms]",
        "ECDF": "ECDF",
        "setup": "setup",
        "best": "beste",
        "worst": "schlechteste",
        "summary": "Übersicht",
        "all_insts": "all Instanzen",
        "all_algos": "all Algorithmen",
        "single_run": "einzelner Run",
        ss.KEY_MEAN_GEOM: "gmean",
        ss.KEY_STDDEV: "sd",
        ss.KEY_MEDIAN: "median",
        ss.KEY_MAXIMUM: "max",
        ss.KEY_MINIMUM: "min",
        ss.KEY_MEAN_ARITH: "mean",
        es.KEY_N_SUCCESS: "#erfolgreich",
    })

    ZH.extend({
        "f": "f",
        "bestF": "f",
        "feasible": "可行的",
        "name": "名称",
        "time": "时间",
        "time_in_ms": "时间(毫秒)",
        "time_in_fes": "时间(目标函数的评价)",
        "lower_bound": "下界",
        "lower_bound_short": "下界",
        bs.F_NAME_NORMALIZED: "归一化f",
        bs.F_NAME_SCALED: "标度f",
        bs.F_NAME_RAW: "f",
        bs.TIME_UNIT_FES: "时间(目标函数的评价)",
        bs.TIME_UNIT_MILLIS: "时间(毫秒)",
        lg.KEY_TOTAL_TIME_MILLIS: "总时间(毫秒)",
        lg.KEY_LAST_IMPROVEMENT_TIME_MILLIS: "最后一次改进是在(毫秒)",
        lg.KEY_TOTAL_FES: "总时间(目标函数的评价)",
        lg.KEY_LAST_IMPROVEMENT_FE: "最后一次改进是在(目标函数的评价)",
        "algorithm_on_instance": "优化算法 \u00d7 优化问题实例",
        "ERT": "经验估计运行时间",
        es.KEY_ERT_FES: "经验估计运行时间[目标函数的评价]",
        es.KEY_ERT_TIME_MILLIS: "经验估计运行时间[毫秒]",
        "ECDF": "经验累积分布函数",
        "setup": "算法配置",
        "best": "最好的",
        "worst": "最糟糕的",
        "summary": "总结",
        "all_insts": "所有优化问题实例",
        "all_algos": "所有的算法",
        "single_run": "一个算法执行",
        ss.KEY_MEAN_GEOM: "几何平均数",
        ss.KEY_STDDEV: "标准偏差",
        ss.KEY_MEDIAN: "中值的",
        ss.KEY_MAXIMUM: "最大限度",
        ss.KEY_MINIMUM: "最低限度",
        ss.KEY_MEAN_ARITH: "算术平均值",
        es.KEY_N_SUCCESS: "成功执行算法的次数",
    })


__lang_setup()  # invoke the language setup
del __lang_setup  # delete the language setup routine



"""A Hybrid EA-FEA Algorithm."""
from collections import Counter
from typing import Callable, Final

from numpy.random import Generator
from pycommons.types import type_error

from moptipy.algorithms.so.ffa.ffa_h import create_h, log_h
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


class EAFEAC(Algorithm1):
    """An implementation of the EAFEA-C."""

    def __init__(self, op0: Op0, op1: Op1, log_h_tbl: bool = False) -> None:
        """
        Create the EAFEA-C.

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param log_h_tbl: should we log the H table?
        """
        super().__init__("eafeaC", op0, op1)
        if not isinstance(log_h_tbl, bool):
            raise type_error(log_h_tbl, "log_h_tbl", bool)
        #: True if we should log the H table, False otherwise
        self.log_h_tbl: Final[bool] = log_h_tbl

    def solve(self, process: Process) -> None:
        """
        Apply the EAFEA-B to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        x_ea = process.create()  # record for current solution of the EA
        x_fea = process.create()  # record for current solution of the FEA
        x_new = process.create()  # record for new solution

        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        should_terminate: Final[Callable] = process.should_terminate
        xcopy: Final[Callable] = process.copy  # copy(dest, source)
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = self.op1.op1  # the unary operator

        h, ofs = create_h(process)  # Allocate the h-table

        # Start at a random point in the search space and evaluate it.
        op0(random, x_ea)  # Create 1 solution randomly and
        y_ea: int | float = evaluate(x_ea) + ofs  # evaluate it.
        xcopy(x_fea, x_ea)  # FEA and EA start with the same initial solution.
        y_fea: int | float = y_ea

        ea_max_no_lt_moves: int = 1  # maximum no-improvement moves for EA
        ea_no_lt_moves: int = 0  # current no-improvement moves
        use_ffa: bool = False  # We start with the EA branch.

        while not should_terminate():  # Until we need to quit...
            # Sample and evaluate new solution.
            op1(random, x_new, x_fea if use_ffa else x_ea)
            y_new: int | float = evaluate(x_new) + ofs
            h[y_new] += 1  # type: ignore  # Always update H.

            if use_ffa:  # The FEA branch uses FFA.
                use_ffa = False  # Always toggle use from FFA to EA.

                h[y_fea] += 1  # type: ignore  # Update H for FEA solution.
                if h[y_new] <= h[y_fea]:  # type: ignore  # FEA acceptance.
                    xcopy(x_ea, x_new)  # Copy solution also to EA.
                    x_fea, x_new = x_new, x_fea
                    y_fea = y_ea = y_new

            else:  # EA or RLS branch performs local search.
                h[y_ea] += 1  # type: ignore  # Update H in *both* branches.

                if y_new <= y_ea:  # The acceptance criterion of RLS / EA.
                    if y_new < y_ea:  # Check if we did an actual improvement.
                        ea_no_lt_moves = 0  # non-improving moves counter = 0.
                        xcopy(x_fea, x_new)  # Copy solution over to FEA.
                        y_fea = y_new  # And store the objective value.
                    else:  # The move was *not* an improvement:
                        ea_no_lt_moves += 1  # Increase non-improved counter.
                    x_ea, x_new = x_new, x_ea  # Accept new solution.
                    y_ea = y_new  # Store objective value.

                if ea_no_lt_moves >= ea_max_no_lt_moves:  # Toggle: EA to FEA.
                    ea_no_lt_moves = 0  # Reset non-improving move counter.
                    ea_max_no_lt_moves += 1  # Increment limit by one.
                    use_ffa = True  # Toggle to FFA.

        if not self.log_h_tbl:
            return  # we are done here

        # After we are done, we want to print the H-table.
        if h[y_ea] == 0:  # type: ignore  # Fix the H-table for the case
            h = Counter()   # that only one FE was performed: In this case,
            h[y_ea] = 1  # make Counter with only a single 1 value inside.

        log_h(process, h, ofs)  # log the H-table



"""Test stuff on signed permutations with repetitions."""

from typing import Any, Callable, Iterable, cast

import numpy as np
from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipy.api.operators import Op0, Op1
from moptipy.spaces.signed_permutations import SignedPermutations
from moptipy.tests.op0 import validate_op0
from moptipy.tests.op1 import validate_op1


def signed_permutations_for_tests(
        perm_filter: Callable[[SignedPermutations], bool] | None = None) \
        -> Iterable[SignedPermutations]:
    """
    Get a sequence of permutations for tests.

    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    :returns: the sequence of SignedPermutations
    """
    r = default_rng()
    pwrs: list[SignedPermutations] = [
        SignedPermutations.standard(2),
        SignedPermutations.standard(3),
        SignedPermutations.standard(4),
        SignedPermutations.standard(5),
        SignedPermutations.standard(6),
        SignedPermutations.standard(12),
        SignedPermutations.standard(23),
        SignedPermutations.with_repetitions(2, 2),
        SignedPermutations.with_repetitions(2, 3),
        SignedPermutations.with_repetitions(3, 2),
        SignedPermutations.with_repetitions(3, 3),
        SignedPermutations.with_repetitions(5, 5),
        SignedPermutations.with_repetitions(int(r.integers(6, 10)),
                                            int(r.integers(2, 7))),
        SignedPermutations.with_repetitions(int(r.integers(2, 5)),
                                            int(r.integers(6, 10))),
        SignedPermutations.with_repetitions(int(r.integers(130, 500)),
                                            int(r.integers(2, 200))),
        SignedPermutations([1, 1, 1, 1, 1, 5, 5, 3]),
        SignedPermutations([2, 1, 1, 1, 1, 1])]
    if perm_filter is not None:
        if not callable(perm_filter):
            raise type_error(perm_filter, "perm_filter", None, call=True)
        pwrs = [p for p in pwrs if perm_filter(p)]
    r.shuffle(cast(list, pwrs))
    return pwrs


def make_signed_permutation_valid(pwr: SignedPermutations) -> \
        Callable[[Generator, np.ndarray], np.ndarray]:
    """
    Create a function that can make permutations with repetitions valid.

    :param pwr: the permutations
    :returns: the function
    """
    def __make_valid(prnd: Generator, x: np.ndarray, ppp=pwr) -> np.ndarray:
        np.copyto(x, ppp.blueprint)
        prnd.shuffle(x)
        x *= ((2 * prnd.integers(low=0, high=2, size=len(x))) - 1)
        return x
    return __make_valid


def validate_op0_on_1_signed_permutations(
        op0: Op0 | Callable[[SignedPermutations], Op0],
        search_space: SignedPermutations,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, SignedPermutations], int] | None = None) -> None:
    """
    Validate the nullary operator on one `SignedPermutations` instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_signed_permutation_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_signed_permutations(
        op0: Op0 | Callable[[SignedPermutations], Op0],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, SignedPermutations], int] | None = None,
        perm_filter: Callable[[SignedPermutations], bool] | None = None) \
        -> None:
    """
    Validate the nullary operator on several `SignedPermutations` instances.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in signed_permutations_for_tests(perm_filter):
        validate_op0_on_1_signed_permutations(op0, pwr, number_of_samples,
                                              min_unique_samples)


def validate_op1_on_1_signed_permutations(
        op1: Op1 | Callable[[SignedPermutations], Op1],
        search_space: SignedPermutations,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, SignedPermutations], int] | None = None) -> None:
    """
    Validate the unary operator on one `SignedPermutations` instance.

    :param op1: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op1": op1(search_space) if callable(op1) else op1,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_signed_permutation_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op1(**args)


def validate_op1_on_signed_permutations(
        op1: Op1 | Callable[[SignedPermutations], Op1],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, SignedPermutations], int] | None = None,
        perm_filter: Callable[[SignedPermutations], bool] | None = None) \
        -> None:
    """
    Validate the unary operator on several `SignedPermutations` instances.

    :param op1: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param perm_filter: an optional filter to sort out permutations we cannot
        use for testing
    """
    for pwr in signed_permutations_for_tests(perm_filter):
        validate_op1_on_1_signed_permutations(op1, pwr, number_of_samples,
                                              min_unique_samples)



"""Test all the example code in the project's README.md file."""

from pycommons.dev.tests.examples_in_md import check_examples_in_md
from pycommons.io.path import file_path


def test_all_examples_from_readme_md() -> None:
    """Test all the example Python codes in the README.md file."""
    # First, we load the README.md file as a single string
    check_examples_in_md(file_path(__file__).up(2).resolve_inside("README.md"))



"""
Processes offer data to both the user and the optimization algorithm.

They provide the information about the optimization process and its current
state as handed to the optimization algorithm and, after the algorithm has
finished, to the user.
This is the multi-objective version of the :class:`~moptipy.api.process.\
Process`-API.
It supports having multiple objective functions.
It also provides a single core objective value, which is the scalarized
result of several objective values.
"""
from typing import Any, Callable, Iterable

import numpy as np

from moptipy.api.mo_archive import MORecord
from moptipy.api.mo_problem import MOProblem
from moptipy.api.process import Process


class MOProcess(MOProblem, Process):
    """
    A multi-objective :class:`~moptipy.api.process.Process` API variant.

    This class encapsulates an optimization process using multiple objective
    functions. It inherits all of its methods from the single-objective
    process :class:`~moptipy.api.process.Process` and extends its API towards
    multi-objective optimization by providing the functionality of the class
    :class:`~moptipy.api.mo_problem.MOProblem`.
    """

    def register(self, x, f: int | float) -> None:
        """Unavailable during multi-objective optimization."""
        raise ValueError(
            "register is not available during multi-objective optimization.")

    def get_archive(self) -> list[MORecord]:
        """
        Get the archive of non-dominated solutions.

        :returns: a list containing all non-dominated solutions currently in
            the archive
        """

    def check_in(self, x: Any, fs: np.ndarray,
                 prune_if_necessary: bool = False) -> bool:
        """
        Check a solution into the archive.

        This method is intended for being invoked after the optimization
        algorithm has finished its work. The algorithm should, by itself,
        maintain the set of interesting solutions during its course. Once
        it has completed all of its computations, it should flush these
        solutions to the process using this method. All the non-dominated
        solutions preserved in the archive will then become available via
        :meth:`get_archive` to the code starting the optimization procedure.

        If you have a sequence of :class:`~moptipy.api.mo_archive.MORecord`
        instances that you want to flush into the archive, you can use the
        convenience method :meth:`~check_in_all` for that purpose.

        :param x: the point in the search space
        :param fs: the vector of objective values
        :param prune_if_necessary: should we prune the archive if it becomes
            too large? `False` means that the archive may grow unbounded
        :returns: `True` if the solution was non-dominated and has actually
            entered the archive, `False` if it has not entered the archive
        """

    def check_in_all(self, recs: Iterable[MORecord],
                     prune_if_necessary: bool = False) -> None:
        """
        Check in all the elements of an `Iterable` of `MORecord` instances.

        This is a convenience wrapper around :meth:`~check_in`.

        :param recs: an iterable sequence of solution + quality vector records
        :param prune_if_necessary: should we prune the archive if it becomes
            too large? `False` means that the archive may grow unbounded
        """
        sci: Callable[[Any, np.ndarray, bool], bool] = self.check_in
        for r in recs:
            sci(r.x, r.fs, prune_if_necessary)

    def get_copy_of_best_fs(self, fs: np.ndarray) -> None:
        """
        Get a copy of the objective vector of the current best solution.

        This always corresponds to the best-so-far solution based on the
        scalarization of the objective vector. It is the best solution that
        the process has seen *so far*, the current best solution.

        You should only call this method if you are either sure that you
        have invoked :meth:`~moptipy.api.process.Process.evaluate`, have
        invoked :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate`, or
        have called :meth:`~moptipy.api.process.Process.has_best` before
        and it returned `True`.

        :param fs: the destination vector to be overwritten

        See Also
            - :meth:`~moptipy.api.process.Process.has_best`
            - :meth:`~moptipy.api.process.Process.get_best_f`
            - :meth:`~moptipy.api.process.Process.get_copy_of_best_x`
            - :meth:`~moptipy.api.process.Process.get_copy_of_best_y`
            - :meth:`~moptipy.api.mo_problem.MOProblem.f_evaluate`
        """

    def __str__(self) -> str:
        """
        Get the name of this process implementation.

        :return: "mo_process"
        """
        return "mo_process"

    def __enter__(self) -> "MOProcess":
        """
        Begin a `with` statement.

        :return: this process itself
        """
        return self



"""The markdown text format driver."""

from io import TextIOBase
from typing import Final

from moptipy.utils.formatted_string import (
    NAN,
    NEGATIVE_INFINITY,
    NUMBER,
    POSITIVE_INFINITY,
    SPECIAL,
    TEXT,
)
from moptipy.utils.latex import SPECIAL_CHARS as __SC
from moptipy.utils.text_format import MODE_TABLE_HEADER, TextFormatDriver

#: the special chars
SPECIAL_CHARS: Final[dict[str, str]] = dict(__SC)
SPECIAL_CHARS["\u2014"] = "&mdash;"


class Markdown(TextFormatDriver):
    r"""
    The markdown text driver.

    >>> from io import StringIO
    >>> from moptipy.utils.formatted_string import FormattedStr
    >>> from moptipy.utils.table import Table
    >>> s = StringIO()
    >>> md = Markdown.instance()
    >>> print(str(md))
    md
    >>> with Table(s, "lrc", md) as t:
    ...     with t.header() as header:
    ...         with header.row() as h:
    ...             h.cell(FormattedStr("1", bold=True))
    ...             h.cell(FormattedStr("2", code=True))
    ...             h.cell(FormattedStr("3", italic=True))
    ...     with t.section() as g:
    ...         with g.row() as r:
    ...             r.cell("a")
    ...             r.cell("b")
    ...             r.cell("c")
    ...         with g.row() as r:
    ...             r.cell("d")
    ...             r.cell("e")
    ...             r.cell("f")
    >>> print(f"'{s.getvalue()}'")
    '|**1**|`2`|*3*|
    |:--|--:|:-:|
    |a|b|c|
    |d|e|f|
    '
    """

    def end_table_header(self, stream: TextIOBase, cols: str) -> None:
        """End the header of a markdown table."""
        for c in cols:
            if c == "l":
                stream.write("|:--")
            elif c == "c":
                stream.write("|:-:")
            elif c == "r":
                stream.write("|--:")
            else:
                raise ValueError(f"Invalid col {c!r} in {cols!r}.")
        stream.write("|\n")

    def begin_table_row(self, stream: TextIOBase, cols: str,
                        section_index: int, row_index: int,
                        row_mode: int) -> None:
        """Begin a row in a markdown table."""
        if (row_mode == MODE_TABLE_HEADER) and (row_index != 0):
            raise ValueError("pandoc markdown only supports one single header"
                             f" row, but encountered row {row_mode + 1}.")

    def end_table_row(self, stream: TextIOBase, cols: str,
                      section_index: int, row_index: int) -> None:
        """End a row in a Markdown table."""
        stream.write("|\n")

    def begin_table_cell(self, stream: TextIOBase, cols: str,
                         section_index: int, row_index: int,
                         col_index: int, cell_mode: int) -> None:
        """Begin a Markdown table cell."""
        stream.write("|")

    def text(self, stream: TextIOBase, text: str, bold: bool, italic: bool,
             code: bool, mode: int) -> None:
        """Print a text string."""
        if len(text) <= 0:
            return
        if bold:
            stream.write("**")
        if italic:
            stream.write("*")
        if code:
            stream.write("`")

        if mode == TEXT:
            stream.write(text)
        elif mode == NUMBER:
            i: int = text.find("e")
            if i < 0:
                i = text.find("E")
            if i > 0:
                stream.write(f"{text[:i]}\\*10^{text[i + 1:]}^")  # \u00D7
            else:
                stream.write(text)
        elif mode == NAN:
            stream.write(r"$\emptyset$")  # \u2205
        elif mode == POSITIVE_INFINITY:
            stream.write(r"$\infty$")  # \u221E
        elif mode == NEGATIVE_INFINITY:
            stream.write(r"$-\infty$")  # -\u221E
        elif mode == SPECIAL:
            s: Final[str] = str(text)
            if s not in SPECIAL_CHARS:
                raise ValueError(f"invalid special character: {s!r}")
            stream.write(SPECIAL_CHARS[s])
        else:
            raise ValueError(f"invalid mode {mode} for text {text!r}.")

        if code:
            stream.write("`")
        if italic:
            stream.write("*")
        if bold:
            stream.write("**")

    def __str__(self):
        """
        Get the appropriate file suffix.

        :returns: the file suffix
        :retval 'md': always
        """
        return "md"

    @staticmethod
    def instance() -> "Markdown":
        """
        Get the markdown format singleton instance.

        :returns: the singleton instance of the Markdown format
        """
        attr: Final[str] = "_instance"
        func: Final = Markdown.instance
        if not hasattr(func, attr):
            setattr(func, attr, Markdown())
        return getattr(func, attr)



"""
The :class:`Lang` class provides facilities for easy internationalization.

The idea here is to have simply tools that provide locale-specific keywords,
texts, and number formats.
"""

from typing import Callable, Final, Iterable

from matplotlib import rc  # type: ignore
from matplotlib.font_manager import (  # type: ignore
    FontProperties,
    findSystemFonts,
)
from pycommons.types import check_int_range, type_error

from moptipy.utils.strings import sanitize_name


class Lang:
    """A language-based dictionary for locale-specific keywords."""

    def __init__(self, name: str, font: str, decimal_stepwidth: int = 3,
                 data: dict[str, str] | None = None,
                 is_default: bool = False):
        """
        Instantiate the language formatter.

        :param name: the short name
        :param font: the font name
        :param decimal_stepwidth: the decimal step width
        :param data: the data
        :param is_default: is this the default language?
        """
        #: the name of the locale
        self.__name: Final[str] = sanitize_name(name)
        if not isinstance(font, str):
            raise type_error(font, "font", str)
        if not isinstance(is_default, bool):
            raise type_error(is_default, "is_default", bool)

        font = font.strip()
        if not font:
            raise ValueError(f"The font cannot be {font!r}.")
        #: the font name
        self.__font: Final[str] = font
        #: the decimal step width
        self.__decimal_stepwidth: Final[int] = check_int_range(
            decimal_stepwidth, "decimal_stepwidth", 1, 10)

        #: the dictionary with the translation data
        self.__dict: Final[dict[str, str]] = {}
        if data:
            self.extend(data)

        #: is this the default language?
        self.__is_default: Final[bool] = is_default

        # register the language
        dc: Final[dict[str, Lang]] = Lang.__get_langs()
        if self.__name in dc:
            raise ValueError(f"Language {self.__name!r} already registered.")
        if is_default:
            for lang in dc.values():
                if lang.__is_default:
                    raise ValueError(
                        f"Language {self.__name!r} cannot be default "
                        f"language, {lang.__name!r} already is!")
        dc[self.__name] = self

        if is_default:
            self.set_current()

    def extend(self, data: dict[str, str]) -> None:
        """
        Add a set of entries to this dictionary.

        :param data: the language-specific data
        """
        if not isinstance(data, dict):
            raise type_error(data, "data", dict)
        for ko, v in data.items():
            k = sanitize_name(ko)
            if k != ko:
                raise ValueError(f"key {ko!r} is different from "
                                 f"its sanitized version {k!r}.")
            if (k in self.__dict) and (self.__dict[k] != v):
                raise ValueError(
                    f"Key {k!r} appears twice, already assigned to "
                    f"{self.__dict[k]!r}, cannot assign to {v!r}.")
            if not isinstance(v, str):
                raise type_error(v, f"value for key {k!r}", str)
            if not v:
                raise ValueError(f"Value for key {k!r} cannot be {v!r}.")
            self.__dict[k] = v

    def filename(self, base: str) -> str:
        """
        Make a suitable filename by appending the language id.

        :param str base: the basename
        :return: the filename
        :rtype: str

        >>> print(Lang.get("en").filename("test"))
        test
        >>> print(Lang.get("zh").filename("test"))
        test_zh
        """
        base = sanitize_name(base)
        return base if self.__is_default else f"{base}_{self.__name}"

    def __repr__(self):
        """
        Get the language's name.

        :return: the language's name
        :rtype: str
        """
        return self.__name

    def __getitem__(self, item: str) -> str:
        """
        Get the language formatting code.

        :param item: the item to get
        :return: the language-specific code
        """
        if not isinstance(item, str):
            raise type_error(item, "item", str)
        return self.__dict[item]

    def format_str(self, item: str, **kwargs) -> str:
        """
        Return a string based on the specified format.

        :param item: the key
        :param kwargs: the keyword-based arguments

        >>> l = Lang.get("en")
        >>> l.extend({"z": "{a}: bla{b}"})
        >>> print(l.format_str("z", a=5, b=6))
        5: bla6
        """
        if not isinstance(item, str):
            raise type_error(item, "item", str)
        fstr: str = self.__dict[item]
        # pylint: disable=W0123 # noqa: DUO104,S307
        return eval(  # nosec # nosemgrep # noqa: DUO104,S307
            f"f{fstr!r}",  # nosec # nosemgrep # noqa: DUO104,B028,S307
            {"__builtins__": None},  # nosec # nosemgrep # noqa:DUO104,S307
            kwargs).strip()  # nosec # nosemgrep # noqa: DUO104,S307

    def font(self) -> str:
        """
        Get the default font for this language.

        :return: the default font for this language

        >>> print(Lang.get("en").font())
        DejaVu Sans
        >>> print(Lang.get("zh").font())
        Noto Sans SC
        """
        return self.__font

    def format_int(self, value: int) -> str:
        """
        Convert an integer to a string.

        :param value: the value
        :returns: a string representation of the value

        >>> print(Lang.get("en").format_int(100000))
        100'000
        >>> print(Lang.get("zh").format_int(100000))
        10'0000
        """
        if not isinstance(value, int):
            raise type_error(value, "value", int)
        if value < 0:
            prefix = "-"
            value = -value
        else:
            prefix = ""

        sss = str(value)
        i = len(sss)
        if i <= self.__decimal_stepwidth:  # no formatting needed
            return prefix + sss

        # We divide the string into equally-sized chunks and insert "'"
        # between them.
        chunks: list[str] = []
        for i in range(i, -1, -self.__decimal_stepwidth):  # noqa
            k: str = sss[i:(i + self.__decimal_stepwidth)]
            if k:
                chunks.insert(0, k)
        if i > 0:
            chunks.insert(0, sss[0:i])
        return prefix + "'".join(chunks)

    @staticmethod
    def __get_langs() -> dict[str, "Lang"]:
        """
        Get the languages decode.

        :return: the languages decode
        """
        att: Final[str] = "__map"
        if not hasattr(Lang.__get_langs, att):
            setattr(Lang.__get_langs, att, {})
        return getattr(Lang.__get_langs, att)

    @staticmethod
    def get(name: str) -> "Lang":
        """
        Get the language of the given key.

        :param name: the language name
        :return: the language
        """
        name = sanitize_name(name)
        lang: Lang | None = Lang.__get_langs().get(name, None)
        if lang:
            return lang
        raise ValueError(f"Unknown language {name!r}.")

    @staticmethod
    def current() -> "Lang":
        """
        Get the current language.

        :return: the current language

        >>> Lang.get("en").set_current()
        >>> print(Lang.current().filename("b"))
        b
        >>> Lang.get("zh").set_current()
        >>> print(Lang.current().filename("b"))
        b_zh
        """
        lang: Final[Lang] = getattr(Lang.__get_langs, "__current")
        if not lang:
            raise ValueError("Huh?")
        return lang

    def set_current(self) -> None:
        """Mark this language as the current one."""
        setattr(Lang.__get_langs, "__current", self)
        rc("font", family=self.font())

    @staticmethod
    def all_langs() -> Iterable["Lang"]:
        """
        Get all presently loaded languages.

        :return: an Iterable of the languages
        """
        val = list(Lang.__get_langs().values())
        val.sort(key=lambda x: x.__name)
        return val

    @staticmethod
    def translate(key: str) -> str:
        """
        Translate the given key to a string in the current language.

        :param key: the key
        :returns: the value of the key in the current language

        >>> EN.extend({'a': 'b'})
        >>> EN.set_current()
        >>> print(Lang.translate("a"))
        b
        >>> DE.extend({'a': 'c'})
        >>> Lang.get("de").set_current()
        >>> print(Lang.translate("a"))
        c
        """
        return Lang.current()[key]

    @staticmethod
    def translate_call(key: str) -> Callable:
        """
        Get a callable that always returns the current translation of a key.

        The callable will ignore all of its parameters and just return the
        translation. This means that you can pass parameters to it and they
        will be ignored.

        :param key: the key to translate
        :returns: the callable doing the translation

        >>> cal = Lang.translate_call("a")
        >>> EN.extend({'a': 'b'})
        >>> EN.set_current()
        >>> print(cal())
        b
        >>> print(cal(1, 2, 3))
        b
        >>> DE.extend({'a': 'c'})
        >>> Lang.get("de").set_current()
        >>> print(cal("x"))
        c
        """
        def __trc(*_, ___key=key) -> str:
            return Lang.current()[___key]
        return __trc

    @staticmethod
    def translate_func(func: str) -> Callable[[str], str]:
        """
        Create a lambda taking a dimensions and presenting a function thereof.

        :param func: the function name
        :returns: the function

        >>> Lang.get("en").set_current()
        >>> Lang.get("en").extend({"ERT": "ERT"})
        >>> Lang.get("en").extend({"FEs": "time in FEs"})
        >>> f = Lang.translate_func("ERT")
        >>> print(f("FEs"))
        ERT\u2009[time in FEs]
        >>> Lang.get("de").set_current()
        >>> Lang.get("de").extend({"ERT": "ERT"})
        >>> Lang.get("de").extend({"FEs": "Zeit in FEs"})
        >>> print(f("FEs"))
        ERT\u2009[Zeit in FEs]
        """
        def __tf(dim: str, f: str = func) -> str:
            return f"{Lang.translate(f)}\u2009[{Lang.translate(dim)}]"
        return __tf


# noinspection PyBroadException
def __get_font(choices: list[str]) -> str:
    """
    Try to find an installed version of the specified font.

    :param choices: the choices of the fonts
    :returns: the installed name of the font, or the value of `choices[0]`
        if no font was found
    """
    if not isinstance(choices, list):
        raise type_error(choices, "choices", list)
    if len(choices) <= 0:
        raise ValueError("no font choices are provided!")
    attr: Final[str] = "fonts_list"
    func: Final = globals()["__get_font"]

    # get the list of installed fonts
    font_list: list[str]
    if not hasattr(func, attr):
        font_list = []
        for fname in findSystemFonts():
            try:
                font_name = FontProperties(
                    fname=fname).get_name().strip()
                if font_name.encode("ascii", "ignore").decode() == font_name:
                    font_list.append(font_name)
            except Exception:  # noqa
                # We can ignore the exceptions here.
                continue  # nosec
        if len(font_list) <= 0:
            raise ValueError("Did not find any font.")
        setattr(func, attr, font_list)
    else:
        font_list = getattr(func, attr)

    # find the installed font
    for choice in choices:
        clc: str = choice.lower()
        found_inside: str | None = None
        found_start: str | None = None
        for got in font_list:
            gotlc = got.lower()
            if clc == gotlc:
                return got
            if gotlc.startswith(clc):
                found_start = got
            if clc in gotlc:
                found_inside = got
        if found_start:
            return found_start
        if found_inside:
            return found_inside

    return choices[0]  # nothing found ... return default


#: the English language
EN: Final[Lang] = Lang("en",
                       __get_font(["DejaVu Sans", "Calibri",
                                   "Arial", "Helvetica"]),
                       3, is_default=True)

#: the German language
DE: Final[Lang] = Lang("de", EN.font(), 3)

#: the Chinese language
ZH: Final[Lang] = Lang("zh",
                       __get_font(["Noto Sans SC", "FangSong", "SimSun",
                                   "Arial Unicode", "SimHei"]),
                       4)

del __get_font  # get rid of no-longer needed data such as fonts list



"""
A binary operator that merges two permutations by adding elements in sequence.

Assume we have two permutations `x0` and `x1`. For each position `i` in the
destination string `dest`, we randomly select the permutation `x` from which
the value should come (so either `x=x0` or `x=x1`). We then store the first
value not yet marked as done from `x` in `dest[i]`, mark that value as
done both in `x0` and `x1`.

This operator may be considered as a generalized version of the Alternating
Position Crossover operator (AP) for the Traveling Salesperson Problem by
Larrañaga et al. (1997). The original AP operator, as described by Larrañaga
et al., simply creates an offspring by selecting alternately the next element
of the fist parent and the next element of the second parent, omitting the
elements already present in the offspring. For example, if `x0` is `12345678`
and `x1` is `37516824`, the AP operator gives the following offspring
`13275468`. Exchanging the parents results in `31725468`.

Our generalized version randomly decides which of the two parent permutations
to use each time, hopefully resulting in a greater variety of possible results.
It also does not skip over a parent if its next element is already used, but
instead picks the next not-yet-used element from that parent.
As mnemonic for the operator, we use `gap`. Larrañaga et al. used `AP` for the
version of the operator that strictly alternates between parents.

1. Pedro Larrañaga, Cindy M. H. Kuijpers, Mikel Poza, and Roberto H. Murga.
   Decomposing Bayesian Networks: Triangulation of the Moral Graph with Genetic
   Algorithms, *Statistics and Computing,* 7(1):19-34, March 1997,
   https://doi.org/10.1023/A:1018553211613
2. Pedro Larrañaga, Cindy M. H. Kuijpers, Roberto H. Murga, Iñaki Inza, and
   S. Dizdarevic. Genetic Algorithms for the Travelling Salesman Problem: A
   Review of Representations and Operators. *Artificial Intelligence Review,*
   13(2):129-170, April 1999. Kluwer Academic Publishers, The Netherlands.
   https://doi.org/10.1023/A:1006529012972
"""
from typing import Final

import numba  # type: ignore
import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.operators import Op2
from moptipy.spaces.permutations import Permutations
from moptipy.utils.nputils import DEFAULT_BOOL


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
# start book
def _op2_gap(r: np.ndarray, dest: np.ndarray,
             x0: np.ndarray, x1: np.ndarray,
             x0_done: np.ndarray, x1_done: np.ndarray) -> None:
    """
    Apply a sequence mix from `x0` and `x1` to `dest`.

    :param r: the random numbers, of length `n - 1` (!!)
    :param dest: the array to receive the result
    :param x0: the first existing point in the search space
    :param x1: the second existing point in the search space
    :param x0_done: a boolean array marking the elements in `x0` that have
        been used
    :param x1_done: a boolean array marking the elements in `x1` that have
        been used
    """
    x0_done.fill(False)  # all values in x0 are available
    x1_done.fill(False)  # all values in x1 are available
    length: Final[int] = len(x0)

    desti: int = 0  # writing to dest starts at index 0
    x0i: int = 0  # first valid value in x0 is at index 0
    x1i: int = 0  # first valid value in x1 is at index 0
    for rr in r:  # repeat until dest is filled, i.e., desti=length
        # randomly chose a source point and pick next operation
        value: int = x0[x0i] if rr == 0 else x1[x1i]
        dest[desti] = value  # store the value in the destination
        desti = desti + 1  # step destination index

        for x0j in range(x0i, length):  # mark value as done in x0
            if (x0[x0j] == value) and (not x0_done[x0j]):  # find
                x0_done[x0j] = True  # value is found and not done
                break  # so we mark it as done and break the loop
        while x0_done[x0i]:  # now we find the next not-yet-done
            x0i = x0i + 1  # value in x0

        for x1j in range(x1i, length):  # mark value as done in x1
            if (x1[x1j] == value) and (not x1_done[x1j]):  # find
                x1_done[x1j] = True  # value is found and not done
                break  # so we mark it as done and break the loop
        while x1_done[x1i]:  # now we find the next not-yet-done
            x1i = x1i + 1  # value in x1

    dest[desti] = x0[x0i]  # = x1[x1i]: the final missing value
# end book


# start book
class Op2GeneralizedAlternatingPosition(Op2):
    """A binary operator trying to preserve the value sequence."""

    def op2(self, random: Generator, dest: np.ndarray,
            x0: np.ndarray, x1: np.ndarray) -> None:
        """
        Apply a sequence mix from `x0` and `x1` to `dest`.

        :param random: the random number generator
        :param dest: the array to receive the result
        :param x0: the first existing point in the search space
        :param x1: the second existing point in the search space
        """
        _op2_gap(random.integers(low=2, high=None, size=len(dest) - 1),
                 dest, x0, x1, self.__x0_done, self.__x1_done)
    # end book

    def __init__(self, space: Permutations) -> None:
        """
        Initialize the GAP crossover operator.

        :param space: the permutation space
        """
        super().__init__()
        if not isinstance(space, Permutations):
            raise type_error(space, "space", Permutations)
        #: the elements that are done in `x0`
        self.__x0_done: Final[np.ndarray] = np.ndarray(
            (space.dimension,), DEFAULT_BOOL)
        #: the elements that are done in `x1`
        self.__x1_done: Final[np.ndarray] = np.ndarray(
            (space.dimension,), DEFAULT_BOOL)

    def __str__(self) -> str:
        """
        Get the name of this binary operator.

        :returns: "gap" for "generalized alternating position crossover",
            the name of this operator
        :retval "gap": always
        """
        return "gap"



"""
An operator swapping several elements in a permutation.

This operator works similarly to
:class:`~moptipy.operators.permutations.op1_swap2.Op1Swap2`, but instead of
swapping two elements (i.e., performing 1 swap), it will perform a random
number of swaps.

First, the operator copies the source string `x` into the destination `dest`.
It then chooses a random index `i1`. It remembers the element at that index
in two variables, `last` and `first`. (`last` will always be the value of the
element will next be overwritten and `first` will never change and remember
the first overwritten element, i.e., the value that we need to store again
into the array in the end.) We also initialize a variable `continue_after`
with `True`. This variable will always tell us if we need to continue and
draw another random index.

We then begin with the main loop, which is iterated as long as
`continue_after` is `True`. Directly as first action in this loop, we set
`continue_after = ri(2) <= 0`. `ri(2)` will return a random integer in
`{0, 1}`. `ri(2) <= 0` will therefore be `True` with probability 0.5.

We now will draw a new random index `i2`. We load the element at index `i2`
into variable `current`. If `current == last`, we immediately go back and
draw another random `i2`. The reason is that we want to store `current` at
`i1`, i.e., where `last` is currently stored. If `current == last`, this
would change nothing. Furthermore, if `continue_after` is `False`, then
`current` must also be different from `first`: If we exit the main loop,
then we will store `first` into the place where we found `current`. Remember,
we will overwrite the element at index `i1` with `current`, so the very first
element we overwrite must eventually be stored back into the array.

OK, if we have obtained an index `i2` whose corresponding element `current`
fulfills all requirements, we can set `dest[i1] = last = current`, i.e.,
remember `current` in `last` and also store it at index `i1`. Next, we will
overwrite the element at index `i2`, so we set `i1 = i2`. If `continue_after`
is `True`, the loop will continue. Otherwise, it will stop.

In the latter case, we store `first` back into the array at the last index
`i1` and do `dest[i1] = first`.

As a result, the operator will swap exactly 2 elements with probability 0.5.
With probability 0.25, it will swap three elements, with 0.125 probability, it
will swap 4 elements, and so on.

1. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
"""
from typing import Callable, Final

import numpy as np
from numpy.random import Generator

from moptipy.api.operators import Op1


# Temporary fix for https://github.com/numba/numba/issues/9103
def swap_n(random: Generator, dest: np.ndarray,  # +book
           x: np.ndarray) -> None:  # +book
    """
    Copy `x` into `dest` and then swap several different values.

    :param random: the random number generator
    :param dest: the array to receive the modified copy of `x`
    :param x: the existing point in the search space

    >>> rand = np.random.default_rng(10)
    >>> xx = np.array(range(10), int)
    >>> out = np.empty(len(xx), int)
    >>> swap_n(rand, out, xx)
    >>> print(out)
    [0 1 7 3 4 5 6 2 8 9]
    >>> swap_n(rand, out, xx)
    >>> print(out)
    [0 1 8 3 4 5 6 7 2 9]
    >>> swap_n(rand, out, xx)
    >>> print(out)
    [0 5 2 3 4 8 6 7 1 9]
    """
    # start book
    dest[:] = x[:]  # First, we copy `x` to `dest`.
    length: Final[int] = len(dest)  # Get the length of `dest`.
    rint: Callable[[int, int], int] = random.integers  # fast call

    i1: int = random.integers(0, length)  # Get the first random index.
    last = first = dest[i1]  # Get the value at the first index.
    continue_after: bool = True  # True -> loop at least once.
    while continue_after:  # Repeat until we should stop
        continue_after = rint(0, 2) <= 0  # 50/50 chance
        while True:  # Loop forever until eligible element found.
            i2: int = rint(0, length)  # new random index.
            current = dest[i2]  # Get the value at the new index.
            if current == last:  # If it is the same as the
                continue  # previous value, continue.
            if continue_after or (current != first):  # If we want
                break  # to stop, then it must be != first value.
        dest[i1] = last = current  # Store value for from i2 at i1.
        i1 = i2  # Update i1 to now point to cell of i2.
    dest[i1] = first  # Finally, store first element back at end.


class Op1SwapN(Op1):
    """
    A unary search operation that swaps several (different) elements.

    It is similar to `swap2`, but instead may perform a random number
    of swaps. After each swap, it decides with probability 0.5 whether
    or not to perform another swap.
    """

    def __init__(self) -> None:
        """Initialize the object."""
        super().__init__()  # -book
        self.op1 = swap_n  # type: ignore  # use function directly
    # end book

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :returns: "swapn", the name of this operator
        :retval "swapn": always
        """
        return "swapn"



"""
Code for testing of implementations of different parts of the moptipy API.

In this package, we provide code that can be used to test different parts of
the moptipy API and implementation. These are not unit tests, but rather code
that can be used to build unit tests. If you want to use moptipy in your code,
then likely you will implement own algorithms and operators. If you want to
test whether they comply with the moptipy specifications, then the functions
here will be helpful.
"""



"""
Approximate the expected running time to reach certain goals.

The (empirically estimated) Expected Running Time (ERT) tries to give an
impression of how long an algorithm needs to reach a certain solution quality.

The ERT for a problem instance is estimated as the ratio of the sum of all FEs
that all the runs consumed until they either have discovered a solution of a
given goal quality or exhausted their budget, divided by the number of runs
that discovered a solution of the goal quality. The ERT is the mean expect
runtime under the assumption of independent restarts after failed runs, which
then may either succeed (consuming the mean runtime of the successful runs) or
fail again (with the observed failure probability, after consuming the
available budget).

The ERT itself can be considered as a function that associates the estimated
runtime given above to all possible solution qualities that can be attained by
an algorithm for a give problem. For qualities/goals that an algorithm did not
attain in any run, the ERT becomes infinite.

1. Kenneth V. Price. Differential Evolution vs. The Functions of the 2nd ICEO.
   In Russ Eberhart, Peter Angeline, Thomas Back, Zbigniew Michalewicz, and
   Xin Yao, editors, *IEEE International Conference on Evolutionary
   Computation,* April 13-16, 1997, Indianapolis, IN, USA, pages 153-157.
   IEEE Computational Intelligence Society. ISBN: 0-7803-3949-5.
   doi: https://doi.org/10.1109/ICEC.1997.592287
2. Nikolaus Hansen, Anne Auger, Steffen Finck, Raymond Ros. *Real-Parameter
   Black-Box Optimization Benchmarking 2010: Experimental Setup.*
   Research Report RR-7215, INRIA. 2010. inria-00462481.
   https://hal.inria.fr/inria-00462481/document/
"""

from dataclasses import dataclass
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable, cast

import numpy as np
from pycommons.io.console import logger
from pycommons.io.csv import COMMENT_START, CSV_SEPARATOR
from pycommons.io.path import Path
from pycommons.strings.string_conv import num_to_str
from pycommons.types import type_error

from moptipy.api.logging import (
    KEY_ALGORITHM,
    KEY_INSTANCE,
)
from moptipy.evaluation._utils import _get_goal_reach_index
from moptipy.evaluation.base import (
    F_NAME_NORMALIZED,
    F_NAME_RAW,
    F_NAME_SCALED,
    KEY_ENCODING,
    KEY_N,
    KEY_OBJECTIVE_FUNCTION,
    MultiRun2DData,
)
from moptipy.evaluation.progress import Progress
from moptipy.utils.logger import (
    KEY_VALUE_SEPARATOR,
)
from moptipy.utils.nputils import DEFAULT_FLOAT, DEFAULT_INT, is_all_finite


def compute_single_ert(source: Iterable[Progress],
                       goal_f: int | float) -> float:
    """
    Compute a single ERT.

    The ERT is the sum of the time that the runs spend with a
    best-so-far quality greater or equal than `goal_f` divided by the
    number of runs that reached `goal_f`. The idea is that the
    unsuccessful runs spent their complete computational budget and
    once they have terminated, we would immediately start a new,
    independent run.

    Warning: `source` must only contain progress objects that contain
    monotonously improving points. It must not contain runs that may get
    worse over time.

    :param source: the source array
    :param goal_f: the goal objective value
    :return: the ERT

    >>> from moptipy.evaluation.progress import Progress as Pr
    >>> from numpy import array as a
    >>> f = "plainF"
    >>> t = "FEs"
    >>> r = [Pr("a", "i", "f", "e", 1, a([1, 4, 8]), t, a([10, 8, 5]), f),
    ...      Pr("a", "i", "f", "e", 2, a([1, 3, 6]), t, a([9, 7, 4]), f),
    ...      Pr("a", "i", "f", "e", 3, a([1, 2, 7, 9]), t, a([8, 7, 6, 3]), f),
    ...      Pr("a", "i", "f", "e", 4, a([1, 12]), t, a([9, 3]), f)]
    >>> print(compute_single_ert(r, 11))
    1.0
    >>> print(compute_single_ert(r, 10))
    1.0
    >>> print(compute_single_ert(r, 9.5))  # (4 + 1 + 1 + 1) / 4 = 1.75
    1.75
    >>> print(compute_single_ert(r, 9))  # (4 + 1 + 1 + 1) / 4 = 1.75
    1.75
    >>> print(compute_single_ert(r, 8.5))  # (4 + 3 + 1 + 12) / 4 = 5
    5.0
    >>> print(compute_single_ert(r, 8))  # (4 + 3 + 1 + 12) / 4 = 5
    5.0
    >>> print(compute_single_ert(r, 7.3))  # (8 + 3 + 2 + 12) / 4 = 6.25
    6.25
    >>> print(compute_single_ert(r, 7))  # (8 + 3 + 2 + 12) / 4 = 6.25
    6.25
    >>> print(compute_single_ert(r, 6.1))  # (8 + 6 + 7 + 12) / 4 = 8.25
    8.25
    >>> print(compute_single_ert(r, 6))  # (8 + 6 + 7 + 12) / 4 = 8.25
    8.25
    >>> print(compute_single_ert(r, 5.7))  # (8 + 6 + 9 + 12) / 4 = 8.75
    8.75
    >>> print(compute_single_ert(r, 5))  # (8 + 6 + 9 + 12) / 4 = 8.75
    8.75
    >>> print(compute_single_ert(r, 4.2))  # (8 + 6 + 9 + 12) / 3 = 11.666...
    11.666666666666666
    >>> print(compute_single_ert(r, 4))  # (8 + 6 + 9 + 12) / 3 = 11.666...
    11.666666666666666
    >>> print(compute_single_ert(r, 3.8))  # (8 + 6 + 9 + 12) / 2 = 17.5
    17.5
    >>> print(compute_single_ert(r, 3))  # (8 + 6 + 9 + 12) / 2 = 17.5
    17.5
    >>> print(compute_single_ert(r, 2.9))
    inf
    >>> print(compute_single_ert(r, 2))
    inf
    """
    n_success: int = 0
    time_sum: int = 0
    for progress in source:
        idx = _get_goal_reach_index(progress.f, goal_f)
        if idx >= 0:
            n_success += 1
        else:
            idx = cast(np.integer, -1)
        time_sum = time_sum + int(progress.time[idx])
    if time_sum <= 0:
        raise ValueError(f"Time sum cannot be {time_sum}.")
    if n_success <= 0:
        return inf
    return time_sum / n_success


@dataclass(frozen=True, init=False, order=False, eq=False)
class Ert(MultiRun2DData):
    """Estimate the Expected Running Time (ERT)."""

    #: The ert function
    ert: np.ndarray

    def __init__(self,
                 algorithm: str | None,
                 instance: str | None,
                 objective: str | None,
                 encoding: str | None,
                 n: int,
                 time_unit: str,
                 f_name: str,
                 ert: np.ndarray):
        """
        Create the Ert function.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param objective: the objective name, if all runs are on the same
            objective function, `None` otherwise
        :param encoding: the encoding name, if all runs are on the same
            encoding and an encoding was actually used, `None` otherwise
        :param n: the total number of runs
        :param time_unit: the time unit
        :param f_name: the objective dimension name
        :param ert: the ert matrix
        """
        super().__init__(algorithm, instance, objective, encoding, n,
                         time_unit, f_name)
        if not isinstance(ert, np.ndarray):
            raise type_error(ert, "ert", np.ndarray)
        ert.flags.writeable = False

        f: Final[np.ndarray] = ert[:, 0]
        if not is_all_finite(f):
            raise ValueError(
                f"Ert x-axis must be all finite, but encountered {f}.")
        ll = f.size
        if (ll > 1) and np.any(f[1:] <= f[:-1]):
            raise ValueError("f data must be strictly increasing,"
                             f"but encountered {f}.")

        t: Final[np.ndarray] = ert[:, 1]
        if np.isfinite(t[0]) or (np.isposinf(t[0])):
            if not is_all_finite(t[1:]):
                raise ValueError(
                    "non-first ert y-axis elements must be all finite.")
            if np.any(t[1:] >= t[:-1]):
                raise ValueError("t data must be strictly decreasing,"
                                 f"but encountered {t}.")
        else:
            raise ValueError(
                f"first ert y-axis element cannot be {t[0]}.")
        object.__setattr__(self, "ert", ert)


def to_csv(ert: Ert, file: str,
           put_header: bool = True) -> Path:
    """
    Store a :class:`Ert` record in a CSV file.

    :param ert: the ERT
    :param file: the file to generate
    :param put_header: should we put a header with meta-data?
    :return: the fully resolved file name
    """
    if not isinstance(ert, Ert):
        raise type_error(ert, "ert", Ert)
    path: Final[Path] = Path(file)
    logger(f"Writing ERT to CSV file {path!r}.")
    path.ensure_parent_dir_exists()

    with path.open_for_write() as out:
        sep: Final[str] = CSV_SEPARATOR
        write: Callable[[str], int] = out.write
        if put_header:
            kv: Final[str] = KEY_VALUE_SEPARATOR
            cmt: Final[str] = COMMENT_START
            if ert.algorithm is not None:
                write(f"{cmt} {KEY_ALGORITHM}{kv}{ert.algorithm}\n")
            if ert.instance is not None:
                write(f"{cmt} {KEY_INSTANCE}{kv}{ert.instance}\n")
            if ert.objective is not None:
                write(f"{cmt} {KEY_OBJECTIVE_FUNCTION}{kv}"
                      f"{ert.objective}\n")
            if ert.encoding is not None:
                write(f"{cmt} {KEY_ENCODING}{kv}{ert.encoding}\n")
            write(f"{cmt} {KEY_N}{kv}{ert.n}\n")
        write(f"{ert.f_name}{sep}ert[{ert.time_unit}]\n")
        for v in ert.ert:
            write(f"{num_to_str(v[0])}{sep}{num_to_str(v[1])}\n")

    logger(f"Done writing ERT to CSV file {path!r}.")
    path.enforce_file()
    return path


def create(source: Iterable[Progress],
           f_lower_bound: int | float | Callable | None = None,
           use_default_lower_bounds: bool = True) -> Ert:
    """
    Create one single Ert record from an iterable of Progress records.

    :param source: the set of progress instances
    :param f_lower_bound: the lower bound for the objective value, or a
        callable that is applied to a progress object to get the lower
        bound
    :param use_default_lower_bounds: should we use the default lower
        bounds
    :return: the Ert record
    """
    if not isinstance(source, Iterable):
        raise type_error(source, "source", Iterable)

    lower_bound: int | float = inf
    if (f_lower_bound is not None) and (not callable(f_lower_bound)):
        if not isfinite(f_lower_bound):
            raise ValueError("f_lower_bound must be finite "
                             f"but is {f_lower_bound}.")
        lower_bound = f_lower_bound
        f_lower_bound = None

    algorithm: str | None = None
    instance: str | None = None
    objective: str | None = None
    encoding: str | None = None
    time_unit: str | None = None
    f_name: str | None = None
    f_list: list[np.ndarray] = []
    n: int = 0

    prgs: Final[list[Progress]] = cast(list[Progress], source) \
        if isinstance(source, list) else list(source)

    for progress in prgs:
        if not isinstance(progress, Progress):
            raise type_error(progress, "progress input", Progress)
        if n <= 0:
            algorithm = progress.algorithm
            instance = progress.instance
            objective = progress.objective
            encoding = progress.encoding
            time_unit = progress.time_unit
            f_name = progress.f_name
        else:
            if algorithm != progress.algorithm:
                algorithm = None
            if instance != progress.instance:
                instance = None
            if objective != progress.objective:
                objective = None
            if encoding != progress.encoding:
                encoding = None
            if time_unit != progress.time_unit:
                raise ValueError(
                    f"Cannot mix time units {time_unit} "
                    f"and {progress.time_unit}.")
            if f_name != progress.f_name:
                raise ValueError(f"Cannot mix f-names {f_name} "
                                 f"and {progress.f_name}.")
        n += 1
        if use_default_lower_bounds and \
                (progress.f_standard is not None) and \
                (progress.f_name == F_NAME_RAW) and \
                (lower_bound > progress.f_standard):
            lower_bound = progress.f_standard
        if f_lower_bound is not None:
            lb = f_lower_bound(progress)
            if not isinstance(lb, int | float):
                raise type_error(lb, "computed lower bound", (int, float))
            if not isfinite(lb):
                raise ValueError(f"Invalid computed lower bound {lb}.")
            lower_bound = min(lower_bound, lb)
        f_list.append(progress.f)

    if n <= 0:
        raise ValueError("Did not encounter any progress information.")

    if use_default_lower_bounds:
        if f_name == F_NAME_SCALED:
            lower_bound = min(lower_bound, 1)
        elif f_name == F_NAME_NORMALIZED:
            lower_bound = min(lower_bound, 0)

    # get unique x-values and make sure that lower bound is included
    has_lb: Final[bool] = isfinite(lower_bound)
    if has_lb:
        f_list.append(np.array([lower_bound]))
    x = np.concatenate(f_list)
    del f_list
    x = np.unique(x)
    if has_lb:
        x = x[x >= lower_bound]
    base_len: Final[int] = x.size

    # prepare for backward iteration over arrays
    indices = np.array([(ppr.f.size - 1) for ppr in prgs],
                       dtype=DEFAULT_INT)
    y = np.empty(base_len, dtype=DEFAULT_FLOAT)

    for out_index in range(base_len):
        f_lim = x[out_index]
        found = 0
        time_sum: int = 0
        for r_idx, pr in enumerate(prgs):
            idx = indices[r_idx]
            f_vals = pr.f
            # did we fulfill the limit?
            if f_vals[idx] <= f_lim:
                found += 1
                # can we move time cursor back?
                if (idx > 0) and (f_vals[idx - 1] <= f_lim):
                    idx -= 1
                    indices[r_idx] = idx
            else:
                # condition not fulfilled, need to use maximum time
                idx = -1
            time_sum = time_sum + int(pr.time[idx])

        # compute ert value: infinite if no run meets condition
        y[out_index] = inf if (found <= 0) else time_sum / found

    # convert the two arrays into one matrix
    ert = np.concatenate((x, y)).reshape((base_len, 2), order="F")

    return Ert(algorithm, instance, objective, encoding, n,
               time_unit, f_name, ert)


def from_progresses(source: Iterable[Progress],
                    consumer: Callable[[Ert], Any],
                    f_lower_bound: float | None = None,
                    use_default_lower_bounds: bool = True,
                    join_all_algorithms: bool = False,
                    join_all_instances: bool = False,
                    join_all_objectives: bool = False,
                    join_all_encodings: bool = False) -> None:
    """
    Compute one or multiple ERTs from a stream of end results.

    :param source: the set of progress instances
    :param f_lower_bound: the lower bound for the objective value
    :param use_default_lower_bounds: should we use the default lower
        bounds
    :param consumer: the destination to which the new records will be
        passed, can be the `append` method of a :class:`list`
    :param join_all_algorithms: should the Ert be aggregated over all
        algorithms
    :param join_all_instances: should the Ert be aggregated over all
        algorithms
    :param join_all_objectives: should the statistics be aggregated over
        all objective functions?
    :param join_all_encodings: should the statistics be aggregated over
        all encodings?
    """
    if not isinstance(source, Iterable):
        raise type_error(source, "source", Iterable)
    if not callable(consumer):
        raise type_error(consumer, "consumer", call=True)
    if not isinstance(join_all_algorithms, bool):
        raise type_error(join_all_algorithms, "join_all_algorithms", bool)
    if not isinstance(join_all_instances, bool):
        raise type_error(join_all_instances, "join_all_instances", bool)

    if (join_all_algorithms and join_all_instances and join_all_objectives
            and join_all_encodings):
        consumer(create(source, f_lower_bound, use_default_lower_bounds))
        return

    sorter: dict[tuple[str, str, str, str], list[Progress]] = {}
    for er in source:
        if not isinstance(er, Progress):
            raise type_error(er, "progress source", Progress)
        key = ("" if join_all_algorithms else er.algorithm,
               "" if join_all_instances else er.instance,
               "" if join_all_objectives else er.objective,
               "" if join_all_encodings else (
                   "" if er.encoding is None else er.encoding))
        if key in sorter:
            lst = sorter[key]
        else:
            lst = []
            sorter[key] = lst
        lst.append(er)

    if len(sorter) <= 0:
        raise ValueError("source must not be empty")

    if len(sorter) > 1:
        keyz = sorted(sorter.keys())
        for key in keyz:
            consumer(create(sorter[key], f_lower_bound,
                            use_default_lower_bounds))
    else:
        consumer(create(next(iter(sorter.values())), f_lower_bound,
                        use_default_lower_bounds))



"""
Convert `moptipy` data to IOHanalyzer data.

The IOHanalyzer (https://iohanalyzer.liacs.nl/) is a tool that can analyze
the performance of iterative optimization heuristics in a wide variety of
ways. It is available both for local installation as well as online for
direct and free use (see, again, https://iohanalyzer.liacs.nl/). The
IOHanalyzer supports many of the diagrams that our evaluation utilities
provide - and several more. Here we provide the function
:func:`moptipy_to_ioh_analyzer` which converts the data generated by the
`moptipy` experimentation function
:func:`~moptipy.api.experiment.run_experiment` to the format that the
IOHanalyzer understands, as documented at
https://iohprofiler.github.io/IOHanalyzer/data/.

Notice that we here have implemented the meta data format version
"0.3.2 and below", as described at https://iohprofiler.github.io/IOHanalyzer\
/data/#iohexperimenter-version-032-and-below.

1. Carola Doerr, Furong Ye, Naama Horesh, Hao Wang, Ofer M. Shir, and Thomas
   Bäck. Benchmarking Discrete Optimization Heuristics with IOHprofiler.
   *Applied Soft Computing* 88(106027):1-21. March 2020.
   doi: https://doi.org/10.1016/j.asoc.2019.106027},
2. Carola Doerr, Hao Wang, Furong Ye, Sander van Rijn, and Thomas Bäck.
   *IOHprofiler: A Benchmarking and Profiling Tool for Iterative Optimization
   Heuristics.* October 15, 2018. New York, NY, USA: Cornell University,
   Cornell Tech. arXiv:1810.05281v1 [cs.NE] 11 Oct 2018.
   https://arxiv.org/pdf/1810.05281.pdf
3. Hao Wang, Diederick Vermetten, Furong Ye, Carola Doerr, and Thomas Bäck.
   IOHanalyzer: Detailed Performance Analyses for Iterative Optimization
   Heuristics. *ACM Transactions on Evolutionary Learning and Optimization*
   2(1)[3]:1-29. March 2022.doi: https://doi.org/10.1145/3510426.
4. Jacob de Nobel and Furong Ye and Diederick Vermetten and Hao Wang and
   Carola Doerr and Thomas Bäck. *IOHexperimenter: Benchmarking Platform for
   Iterative Optimization Heuristics.* November 2021. New York, NY, USA:
   Cornell University, Cornell Tech. arXiv:2111.04077v2 [cs.NE] 17 Apr 2022.
   https://arxiv.org/pdf/2111.04077.pdf
5. Data Format: Iterative Optimization Heuristics Profiler.
   https://iohprofiler.github.io/IOHanalyzer/data/
"""

import argparse
import contextlib
from typing import Any, Callable, Final

import numpy as np
from pycommons.io.console import logger
from pycommons.io.path import Path, directory_path
from pycommons.strings.string_conv import float_to_str
from pycommons.types import check_int_range, type_error

from moptipy.evaluation.base import F_NAME_RAW, TIME_UNIT_FES, check_f_name
from moptipy.evaluation.progress import Progress, from_logs
from moptipy.utils.help import moptipy_argparser


def __prefix(s: str) -> str:
    """
    Return `xxx` if `s` is of the form `xxx_i` and `i` is `int`.

    :param s: the function name
    :return: the dimension
    """
    idx = s.rfind("_")
    if idx > 0:
        with contextlib.suppress(ValueError):
            i = int(s[idx + 1:])
            if i > 0:
                return s[:idx].strip()
    return s


def __int_suffix(s: str) -> int:
    """
    Return `i` if `s` is of the form `xxx_i` and `i` is `int`.

    This function tries to check if the name

    :param s: the function name
    :return: the dimension
    """
    idx = s.rfind("_")
    if idx > 0:
        with contextlib.suppress(ValueError):
            i = int(s[idx + 1:])
            if i > 0:
                return i
    return 1


def __npstr(a: Any) -> str:
    """
    Convert numpy numbers to strings.

    :param a: the input
    :returns: a string
    """
    return str(int(a)) if isinstance(a, np.integer) \
        else float_to_str(float(a))


def moptipy_to_ioh_analyzer(
        results_dir: str,
        dest_dir: str,
        inst_name_to_func_id: Callable[[str], str] = __prefix,
        inst_name_to_dimension: Callable[[str], int] = __int_suffix,
        inst_name_to_inst_id: Callable[[str], int] = lambda x: 1,
        suite: str = "moptipy",
        f_name: str = F_NAME_RAW,
        f_standard: dict[str, int | float] | None = None) -> None:
    """
    Convert moptipy log data to IOHanalyzer log data.

    :param results_dir: the directory where we can find the results in moptipy
        format
    :param dest_dir: the directory where we would write the IOHanalyzer style
        data
    :param inst_name_to_func_id: convert the instance name to a function ID
    :param inst_name_to_dimension: convert an instance name to a function
        dimension
    :param inst_name_to_inst_id: convert the instance name an instance ID,
        which must be a positive integer number
    :param suite: the suite name
    :param f_name: the objective name
    :param f_standard: a dictionary mapping instances to standard values
    """
    source: Final[Path] = directory_path(results_dir)
    dest: Final[Path] = Path(dest_dir)
    dest.ensure_dir_exists()
    logger(f"converting the moptipy log files in {source!r} to "
           f"IOHprofiler data in {dest!r}. First we load the data.")

    if (f_standard is not None) and (not isinstance(f_standard, dict)):
        raise type_error(f_standard, "f_standard", dict)
    if not isinstance(suite, str):
        raise type_error(suite, "suite", str)
    if (len(suite) <= 0) or (" " in suite):
        raise ValueError(f"invalid suite name {suite!r}")
    if not callable(inst_name_to_func_id):
        raise type_error(
            inst_name_to_func_id, "inst_name_to_func_id", call=True)
    if not callable(inst_name_to_dimension):
        raise type_error(
            inst_name_to_dimension, "inst_name_to_dimension", call=True)
    if not callable(inst_name_to_inst_id):
        raise type_error(
            inst_name_to_inst_id, "inst_name_to_inst_id", call=True)

    # the data
    data: Final[dict[str, dict[str, dict[int, list[
        tuple[int, np.ndarray, np.ndarray]]]]]] = {}

    # this consumer collects all the data in a structured fashion
    def __consume(progress: Progress) -> None:
        nonlocal data  # noqa
        nonlocal inst_name_to_func_id
        nonlocal inst_name_to_dimension
        nonlocal inst_name_to_inst_id

        _algo: dict[str, dict[int, list[tuple[int, np.ndarray, np.ndarray]]]]
        if progress.algorithm in data:
            _algo = data[progress.algorithm]
        else:
            data[progress.algorithm] = _algo = {}
        _func_id: Final[str] = inst_name_to_func_id(progress.instance)
        if not isinstance(_func_id, str):
            raise type_error(_func_id, "function id", str)
        if (len(_func_id) <= 0) or ("_" in _func_id):
            raise ValueError(f"invalid function id {_func_id!r}.")
        _func: dict[int, list[tuple[int, np.ndarray, np.ndarray]]]
        if _func_id in _algo:
            _func = _algo[_func_id]
        else:
            _algo[_func_id] = _func = {}
        _dim: Final[int] = check_int_range(
            inst_name_to_dimension(progress.instance), "dimension", 1)
        _iid: Final[int] = check_int_range(
            inst_name_to_inst_id(progress.instance), "instance id", 1)
        _res: Final[tuple[int, np.ndarray, np.ndarray]] = \
            (_iid, progress.time, progress.f)
        if _dim in _func:
            _func[_dim].append(_res)
        else:
            _func[_dim] = [_res]

    from_logs(source, consumer=__consume, time_unit=TIME_UNIT_FES,
              f_name=check_f_name(f_name), f_standard=f_standard,
              only_improvements=True)

    if len(data) <= 0:
        raise ValueError("did not find any data!")
    logger(f"finished loading data from {len(data)} algorithms, "
           "now writing output.")

    for algo_name in sorted(data.keys()):
        algo = data[algo_name]
        algo_dir: Path = dest.resolve_inside(algo_name)
        algo_dir.ensure_dir_exists()
        logger(f"writing output for {len(algo)} functions of "
               f"algorithm {algo_name!r}.")
        for func_id in sorted(algo.keys()):
            func_dir: Path = algo_dir.resolve_inside(f"data_f{func_id}")
            func_dir.ensure_dir_exists()
            func = algo[func_id]
            logger(f"writing output for algorithm {algo_name!r} and "
                   f"function {func_id!r}, got {len(func)} dimensions.")

            func_name = f"IOHprofiler_f{func_id}"
            with algo_dir.resolve_inside(
                    f"{func_name}.info").open_for_write() as info:
                for dimi in sorted(func.keys()):
                    dim_path = func_dir.resolve_inside(
                        f"{func_name}_DIM{dimi}.dat")
                    info.write(f"suite = {suite!r}, funcId = {func_id!r}, "
                               f"DIM = {dimi}, algId = {algo_name!r}\n")
                    info.write("%\n")
                    info.write(dim_path[len(algo_dir) + 1:])
                    with dim_path.open_for_write() as dat:
                        for per_dim in sorted(
                                func[dimi], key=lambda x:
                                (x[0], x[2][-1], x[1][-1])):
                            info.write(f", {per_dim[0]}:")
                            fes = per_dim[1]
                            f = per_dim[2]
                            info.write(__npstr(fes[-1]))
                            info.write("|")
                            info.write(__npstr(f[-1]))
                            dat.write(
                                '"function evaluation" "best-so-far f(x)"\n')
                            for i, ff in enumerate(f):
                                dat.write(
                                    f"{__npstr(fes[i])} {__npstr(ff)}\n")
                        dat.write("\n")
                info.write("\n")
    del data
    logger("finished converting moptipy data to IOHprofiler data.")


# Run conversion if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipy_argparser(
        __file__,
        "Convert experimental results from the moptipy to the "
        "IOHanalyzer format.",
        "The experiment execution API of moptipy creates an output "
        "folder structure with clearly specified log files that can be"
        " evaluated with our experimental data analysis API. The "
        "IOHprofiler tool chain offers another format (specified in "
        "https://iohprofiler.github.io/IOHanalyzer/data/). With this "
        "tool here, you can convert from the moptipy to the "
        "IOHprofiler format.")
    parser.add_argument(
        "source", help="the directory with moptipy log files", type=Path,
        nargs="?", default="./results")
    parser.add_argument(
        "dest", help="the directory to write the IOHanalyzer data to",
        type=Path, nargs="?", default="./IOHanalyzer")
    args: Final[argparse.Namespace] = parser.parse_args()
    moptipy_to_ioh_analyzer(args.source, args.dest)



"""
The implementation of the Randomized Local Search algorithm `rls`.

The algorithm starts by applying the nullary search operator, an
implementation of :meth:`~moptipy.api.operators.Op0.op0`, to sample
one fully random solution. This is the first best-so-far solution.
In each step, it applies the unary operator, an implementation of
:meth:`~moptipy.api.operators.Op1.op1`, to the best-so-far solution to
obtain a new, similar solution. If this new solution is not worse than
the current best-so-far solution, it replaces this solution. Otherwise,
it is discarded.

The `rls` algorithm is a simple local search that accepts all
non-deteriorating moves. It is thus similar to the simple hill climber
`hc` implemented in
:class:`~moptipy.algorithms.so.hill_climber.HillClimber`, which, however,
accepts strictly improving moves. `rls` is also equivalent to a
`(mu+lambda)`-EA without crossover as implemented in
:class:`~moptipy.algorithms.so.ea.EA` if the same unary and nullary operator
are used and `mu=1`, `lambda=1`, and `br=0`. `rls`, however, will be
faster as it does not represent a population of solutions as list of
objects but can directly utilize local variables.

Strictly speaking, the name "Randomized Local Search" only fits
partially to the algorithm we implement here. Take the discrete search
domain, where the search spaces are bit strings of a fixed length `n`,
as an example. The name "Randomized Local Search" and the abbreviation
`rls` has a fixed meaning on this domain: It is the algorithm that
starts with a random solution and flips exactly one randomly chosen bit
in each step. This corresponds to our `rls` algorithm with the operator
:class:`~moptipy.operators.bitstrings.op1_flip1.Op1Flip1`. However,
an algorithm that starts with a random solution and flips a number of
bits sampled from a Binomial distribution is called `(1+1) EA`. Now
this algorithm corresponds again to our `rls` algorithm, but this time
with operator
:class:`~moptipy.operators.bitstrings.op1_m_over_n_flip.Op1MoverNflip`.
In other words, we can implement (at least) two algorithms with
well-known and fixed names by plugging different operators into our
`rls` approach. One of them is called `RLS`, the other one is called
`(1+1) EA`. Now this is somewhat confusing but results from the general
nature of our basic framework. Regardless of what we do, we will have
some form of name clash here. We advise the user of our algorithms to
be careful with respect to literature and scientific conventions when
using our framework.

1. Frank Neumann and Ingo Wegener. Randomized Local Search, Evolutionary
   Algorithms, and the Minimum Spanning Tree Problem. *Theoretical Computer
   Science.* 378(1):32-40, June 2007.
   https://doi.org/10.1016/j.tcs.2006.11.002,
   https://eldorado.tu-dortmund.de/bitstream/2003/5454/1/165.pdf
2. Holger H. Hoos and Thomas Stützle. *Stochastic Local Search: Foundations
   and Applications.* 2005. ISBN: 1493303732. In The Morgan Kaufmann Series in
   Artificial Intelligence. Amsterdam, The Netherlands: Elsevier.
3. Thomas Weise. *Optimization Algorithms.* 2021. Hefei, Anhui, China:
   Institute of Applied Optimization (IAO), School of Artificial Intelligence
   and Big Data, Hefei University. http://thomasweise.github.io/oa/
"""
from typing import Callable, Final

from numpy.random import Generator

from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1
from moptipy.api.process import Process


# start book
class RLS(Algorithm1):
    """
    The RLS is a simple local search accepting all non-worsening moves.

    In each step, an RLS creates a modified copy `new_x` of the
    current best solution `best_x`. If `new_x` is not worse than `best_x`,
    it becomes the new `best_x`. Otherwise, it is discarded.
    """

    def solve(self, process: Process) -> None:
        """
        Apply the RLS to an optimization problem.

        :param process: the black-box process object
        """
        # Create records for old and new point in the search space.
        best_x = process.create()  # record for best-so-far solution
        new_x = process.create()  # record for new solution
        # Obtain the random number generator.
        random: Final[Generator] = process.get_random()

        # Put function references in variables to save time.
        evaluate: Final[Callable] = process.evaluate  # the objective
        op1: Final[Callable] = self.op1.op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate

        # Start at a random point in the search space and evaluate it.
        self.op0.op0(random, best_x)  # Create 1 solution randomly and
        best_f: int | float = evaluate(best_x)  # evaluate it.

        while not should_terminate():  # Until we need to quit...
            op1(random, new_x, best_x)  # new_x = neighbor of best_x
            new_f: int | float = evaluate(new_x)
            if new_f <= best_f:  # new_x is not worse than best_x?
                best_f = new_f  # Store its objective value.
                best_x, new_x = new_x, best_x  # Swap best and new.

    # end book

    def __init__(self, op0: Op0, op1: Op1) -> None:
        """
        Create the randomized local search (rls).

        :param op0: the nullary search operator
        :param op1: the unary search operator
        """
        super().__init__("rls", op0, op1)
